# import feature_len as fl
# from . import utils  as ul
                                                                                                                                                                                                    

def install(path):
    import os
    import zipfile
    from urllib import request

    jar = path / "ChromHMM/ChromHMM.jar" 
    if jar.exists():
        print("ChromHMM has installed!")
        exit()

    ChromHMM_url = "http://compbio.mit.edu/ChromHMM/ChromHMM.zip"
    software = path / "ChromHMM.zip"
    with request.urlopen(ChromHMM_url) as f:
        data = f.read()
        with open(software, "wb") as h:
            h.write(data)
    with zipfile.ZipFile(software, 'r') as zip_ref:
        zip_ref.extractall(path)
    os.remove(software)
    







class EventID:

    def __init__(self, event_id):
        self.event_id = event_id
        self.parse_id(event_id)
        self.parse_AS_coor()

    def parse_id(self, event_id):
        gene_id, tmp = event_id.split(";")
        AS_type, Chr, *coordinates, strand = tmp.split(":")
        self.gene_id = gene_id
        self.AS_type = AS_type
        self.Chr = Chr
        self.strand = strand
        self.coordinate_str = "-".join(coordinates)

    @property
    def alter_element_len(self):
        alter_len = self.alter_element_coor[1] - self.alter_element_coor[0]
        return alter_len

    def parse_AS_coor(self): 

        AS_func_dic = {
            "SE": self._parse_SE_coor,
            "MX": self._parse_MX_coor,
            "A5": self._parse_SS_coor,
            "A3": self._parse_SS_coor,
            "RI": self._parse_RI_coor,
            "AF": self._parse_FL_coor,
            "AL": self._parse_FL_coor
        }
        self.coordinates = AS_func_dic[self.AS_type]()

    def _parse_SE_coor(self): 
        nums = self.coordinate_str.split("-")
        e1, s2, e2, s3 = map(int, nums)
        self.alter_element_coor = (s2, e2)
        return (e1, s2, e2, s3)
    
    def _parse_MX_coor(self):
        nums = self.coordinate_str.split("-")
        e1, s2, e2, s4, _, s3, e3, _ = map(int, nums)
        self.alter_element_coor = (s2, e2)
        return (e1, s2, e2, s3, e3, s4)

    def _parse_SS_coor(self):
        nums = self.coordinate_str.split("-")
        A5_fs = all([self.strand == "+", self.AS_type == "A5"])
        A3_rs = all([self.strand == "-", self.AS_type == "A3"])

        A5_rs = all([self.strand == "-", self.AS_type == "A5"])
        A3_fs = all([self.strand == "+", self.AS_type == "A3"])

        if A5_fs or A3_rs:
            e2, s3, e1, _ = map(int, nums)
            coors = (e1, e2, s3)
            self.alter_element_coor = (e1, e2)
        elif A5_rs or A3_fs:
            e1, s2, _, s3 = map(int, nums)
            coors = (e1, s2, s3)
            self.alter_element_coor = (s2, s3)
        else:
            coors = ()
        return coors

    def _parse_RI_coor(self):
        nums = self.coordinate_str.split("-")
        s1, e1, s2, e2 = map(int, nums)
        self.alter_element_coor = (e1, s2)
        return (s1, e1, s2, e2)

    def _parse_FL_coor(self):
        nums = self.coordinate_str.split("-")
        AF_fs = all([self.strand == "+", self.AS_type == "AF"])
        AL_rs = all([self.strand == "-", self.AS_type == "AL"])

        AF_rs = all([self.strand == "-", self.AS_type == "AF"])
        AL_fs = all([self.strand == "+", self.AS_type == "AL"])

        if AF_fs or AL_rs:
            s1, e1, s3, s2, e2, _ = map(int, nums)
            coors = (s1, e1, s2, e2, s3)
            self.alter_element_coor = (s1, e1)
        elif AF_rs or AL_fs:
            e1, s2, e2, _, s3, e3 = map(int, nums)
            coors = (e1, s2, e2, s3, e3)
            self.alter_element_coor = (s3, e3)
        else:
            coors = ()
        return coors


##### they are old function for length parsing

def SE_exon_len(event_id):
    start, end = event_id.split("-")[1].split(":")
    return int(end) - int(start)

def MX_exon_len(event_id):
    _, exon1, _, exon2 = event_id.split("-")[:4]
    s1, e1 = exon1.split(":")
    s2, e2 = exon2.split(":")
    len1 = int(e1) - int(s1)
    len2 = int(e2) - int(s2)
    return len1, len2

def A5_exon_len(event_id):
    splice1, splice2 = event_id.split(":")[2: 4]
    if event_id.endswith("+"):
        end = int(splice1.split("-")[0])
        start = int(splice2.split("-")[0])
    elif event_id.endswith("-"):
         end = int(splice2.split("-")[1])
         start = int(splice1.split("-")[1])
    return end - start

def A3_exon_len(event_id):
    splice1, splice2 = event_id.split(":")[2: 4]
    if event_id.endswith("+"):
        end = int(splice2.split("-")[1])
        start = int(splice1.split("-")[1])
    elif event_id.endswith("-"):
         end = int(splice1.split("-")[0])
         start = int(splice2.split("-")[0])
    return end - start

def RI_intron_len(event_id):
    start, end = event_id.split(":")[3].split("-")
    return int(end) - int(start)

def AF_exon_len(event_id):
    if event_id.endswith("+"):
        s1, e1 = event_id.split("-")[0].split(":")[-2:]
        s2, e2 = event_id.split("-")[1].split(":")[1:]
    elif event_id.endswith("-"):
        s1, e1 = event_id.split("-")[-2].split(":")[:2]
        s2, e2 = event_id.split("-")[1].split(":")[:2]
    len1 = int(e1) - int(s1)
    len2 = int(e2) - int(s2)
    return len1, len2

def AL_exon_len(event_id):
    if event_id.endswith("+"):
        s1, e1 = event_id.split("-")[2].split(":")[:2]
        s2, e2 = event_id.split("-")[1].split(":")[:2]
    elif event_id.endswith("-"):
        s1, e1 = event_id.split("-")[0].split(":")[-2:]
        s2, e2 = event_id.split("-")[1].split(":")[1:]
    len1 = int(e1) - int(s1)
    len2 = int(e2) - int(s2)
    return len1, len2

def AS_len(event_id):
    as_type = event_id.split(";")[1][:2]
    len_func_dic = {
        "SE": SE_exon_len,
        "MX": MX_exon_len,
        "A5": A5_exon_len,
        "A3": A3_exon_len,
        "RI": RI_intron_len,
        "AF": AF_exon_len,
        "AL": AL_exon_len
    }
    func = len_func_dic[as_type]
    f_len = func(event_id)
    if isinstance(f_len, tuple) and len(f_len) == 2:
        res = {"len": f_len[0], "len2":  f_len[1]}
    elif isinstance(f_len, int):
        res = {"len": f_len}
    return res


if "__main__" == "__name__":
    se_test = "ENSMUSG00000024169.17;SE:chr17:25244920-25247778:25247953-25251059:+"
    mx_test = "ENSMUSG00000000037.18;MX:chrX:160004745-160006153:160006236-160008873:160004745-160007457:160007540-160008873:+"

    a5_test_f = "ENSMUSG00000000134.18;A5:chrX:7629452-7631746:7629299-7631746:+"
    a5_test_r = "ENSMUSG00000000131.16;A5:chr7:125708072-125709086:125708072-125709126:-"

    a3_test_f = "ENSMUSG00000000127.16;A3:chr17:64280152-64288475:64280152-64288478:+"
    a3_test_r = "ENSMUSG00000000131.16;A3:chr7:125704106-125704273:125703961-125704273:-"

    ri_test = "ENSMUSG00000000031.17;RI:chr7:142129268:142129851-142129940:142130267:-"

    af_test_f = "ENSMUSG00000000049.12;AF:chr11:108234180:108234350-108286128:108271990:108272222-108286128:+"
    af_test_r = "ENSMUSG00000000103.13;AF:chrY:2150121-2150216:2150346:2150121-2170149:2170409:-"

    al_test_f = "ENSMUSG00000000037.18;AL:chrX:160029363-160038026:160038655:160029363-160039562:160039882:+"
    al_test_r = "ENSMUSG00000000384.16;AL:chr11:6570002:6570143-6570742:6570180:6570235-6570742:-"

    a = EventID(al_test_r)
    print(a.Chr, a.gene_id, a.coordinates, a.strand, a.alter_element_len)

    # a = SE_exon_len(se_test)
    # b = MX_exon_len(mx_test)
    # c_f = A5_exon_len(a5_test_f)
    # c_r = A5_exon_len(a5_test_r)

    # d_f = A3_exon_len(a3_test_f)
    # d_r = A3_exon_len(a3_test_r)

    # e = RI_intron_len(ri_test)

    # f_f = AF_exon_len(af_test_f)
    # f_r = AF_exon_len(af_test_r)

    # g_f = AL_exon_len(al_test_f)
    # g_r = AL_exon_len(al_test_r)

    # print(a)
    # print(b)
    # print(c_f, c_r)
    # print(d_f, d_r)
    # print(e)

    # print(f_f)
    # print(f_r)

    # print(g_f, g_r)


"""
Functions and Classes to parse and store GTF file data


@author: Miha Skalic
@email: miha.skalic[at]gmail.com
"""

import sys
import os


class Genome():
    def __init__(self):
        self.data = {}

    def add_to_genes(self, t_dict):
        """
        Appends found exon to his place in genome
        """
        # and strand if and chr if not present
        if t_dict['chr'] not in self.data:
            self.data[t_dict['chr']] = {}
        if t_dict['strand'] not in self.data[t_dict['chr']]:
            self.data[t_dict['chr']][t_dict['strand']] = {}

        # append to gene:
        my_gene = self.data[t_dict['chr']][t_dict['strand']].get(t_dict['gene'], Gene(t_dict['gene']))
        my_gene.add_transcript(t_dict)
        self.data[t_dict['chr']][t_dict['strand']][t_dict['gene']] = my_gene

    def __iter__(self):
        """
        Generates tuples: Gene object, chr str, strand str
        """
        for chr_p in self.data:
            for strand in self.data[chr_p]:
                for gene in self.data[chr_p][strand]:
                    yield self.data[chr_p][strand][gene], chr_p, strand

    def sort_transcripts(self):
        """
        For each gene in genome sorts transcripts (for further
        precessing).
        """
        for gene, _, _ in self.__iter__():
            gene.sort_transcripts()

    def split_genes(self):
        """
        Checks Transkript overlap and split gene if transcripts do not overlap
        """
        # new list consists of pairs (chr, strand, gene_name, list of first transcript indencies)
        new_genes = []

        # construct new genes
        for gene, xchr, strand in self.__iter__():
            overlaps = gene.get_non_overlaps()
            if overlaps:
                new_gene_group = gene.split_gene(overlaps)
                new_genes.append((xchr, strand, gene.name, new_gene_group))

        # delete old gene and add new genes
        for xchr, strand, gname, xgenes in new_genes:
            self.data[xchr][strand].pop(gname)
            for ngene in xgenes:
                self.data[xchr][strand][ngene.name] = ngene

    def merge_two_genes(self, gene1, gene2, chr_p, strand):
        """
        deletes the old gene and creates new merged gene
        """
        # construct new gene
        merge_g = Gene('{}_and_{}'.format(gene1.name, gene2.name))
        merge_g.transcripts = gene1.transcripts
        merge_g.transcripts.update(gene2.transcripts)
        merge_g.sort_transcripts()

        # add new genes old gene
        self.data[chr_p][strand][merge_g.name] = merge_g

        # delte old genes
        self.data[chr_p][strand].pop(gene1.name)
        self.data[chr_p][strand].pop(gene2.name)
        span = merge_g.get_span()
        return span[0], span[1], merge_g

    def merge_overlaps(self, g_list, chr_p, strand):
        """
        Given a ordered list of genes creates new polled genes.
        """
        current_g = g_list.pop(0)
        while g_list:
            new_g = g_list.pop(0)
            # check if they overlap
            if new_g[0] <= current_g[1]:
                current_g = self.merge_two_genes(current_g[2], new_g[2], chr_p, strand)
            else:
                current_g = new_g

    def poll_genes(self):
        """
        Pools genes that overlap
        """

        for chr_p in self.data:
            for strand in self.data[chr_p]:
                # create list of genes
                g_list = []
                for gene in self.data[chr_p][strand]:
                    coords = self.data[chr_p][strand][gene].get_span()
                    g_list.append((coords[0], coords[1], self.data[chr_p][strand][gene]))
                g_list.sort()
                self.merge_overlaps(g_list, chr_p, strand)


class Gene():
    def __init__(self, name):
        self.name = name
        self.transcripts = {}
        self.sortedTranscripts = []

    def __lt__(self, other):
        return 0

    def add_transcript(self, t_dict):
        transcript = self.transcripts.get(t_dict['transcript'], Transcript())
        transcript.add_exon(t_dict['coordinates'][0], t_dict['coordinates'][1])
        self.transcripts[t_dict['transcript']] = transcript

    def sort_transcripts(self):
        self.sortedTranscripts = sorted(
            self.transcripts.keys(),
            key=lambda x: self.transcripts[x].span)

    def get_non_overlaps(self):
        """
        Returns indencies of transcripts that do not overlap with previous transcript
        """
        new_sites = []
        m_upper = self.transcripts[self.sortedTranscripts[0]].span[1]
        for i in range(1, len(self.sortedTranscripts)):
            if self.transcripts[self.sortedTranscripts[i]].span[0] > m_upper:
                    # self.transcripts[self.sortedTranscripts[i - 1]].span[1]:
                new_sites.append(i)
            m_upper = max([m_upper, self.transcripts[self.sortedTranscripts[i]].span[1]])
        return new_sites

    def split_gene(self, idxs):
        """
        Splits gene into two new genes at give incendies.
        Returns two new gene objects
        """
        new_genes = []
        idxs = [0] + idxs + [len(self.sortedTranscripts)]
        for i in range(1, len(idxs)):
            gname = '{}_locus{}'.format(self.name, i)
            new_gene = Gene(gname)
            new_gene.sortedTranscripts = self.sortedTranscripts[idxs[i-1]:idxs[i]]
            new_gene.transcripts = {xtranscript: self.transcripts[xtranscript]
                                    for xtranscript in new_gene.sortedTranscripts}
            new_genes.append(new_gene)
        return new_genes

    def get_span(self):
        """
        Fetches the gene range span and returns results in tuple
        """
        return min([self.transcripts[transc].span[0] for transc in self.transcripts]), \
            max([self.transcripts[transc].span[1] for transc in self.transcripts])


class Transcript():
    def __init__(self):
        self.exons = set()
        self.span = (float('inf'), float('-inf'))

    def add_exon(self, start, stop):
        """
        Adds exon to transkript
        """
        self.exons.add((start, stop))
        self.span = (min([self.span[0], start]), max([self.span[1], stop]))


def parse_gtf_line(line, l_count, logger):
    #  Comment line
    if line.startswith('#'):
        return
    line = line.strip().split('\t')

    # missformated line
    if len(line) != 9:
        logger.info('Missmatch in number of Fields. skipping line {}'.format(l_count))
        return
    # Skip if it not an exon
    if line[2] != 'exon':
        return
    # parsing the attributes
    attributes = [att.split(' ', 1) for att in line[-1].strip(';').split('; ')]
    #Some gtf has an extra space in column 9 (like GRCh37 v67 Ensembl). Remove this space if it's in there and reformat the attribute
    if(attributes[0][0]==""):
        attributes[0] = attributes[0][1].split(" ")
    att_dict = dict(map(lambda x: (x[0], x[1].strip('"')), attributes))
    if 'gene_id' not in att_dict or 'transcript_id' not in att_dict:
        logger.info('Missing gene_id or transcript_id field in line {}'.format(l_count))
        return

    return {'gene': att_dict['gene_id'],
            'transcript': att_dict['transcript_id'],
            'coordinates': (int(line[3]), int(line[4])),
            'strand': line[6],
            'chr': line[0]}


def gtf_reader(parse_file, logger):
    """
    Parses the input GTF file and creates exon
    representations for further processing
    """
    l_count = 0
    all_results = []
    # Check if file exists
    if not os.path.isfile(parse_file):
        sys.stderr.write("Input file does not exist. Quiting\n")
        exit(1)
    # logger.info("Reading input data.")
    with open(parse_file, 'r') as handle:
        for line in handle:
            l_count += 1
            results = parse_gtf_line(line, l_count, logger)
            if results:
                all_results.append(results)
    return all_results



# -*- coding: utf-8 -*-
"""
Created on Wed May 25 04:20:00 CEST 2016

@authors: Juan C Entizne
@email: juancarlos.entizne01[at]estudiant.upf.edu

Modified by Juan L. Trincado
@email: juanluis.trincado[at].upf.edu

"""

import os
import logging
from lib.diff_tools import multiple_conditions_analysis
from argparse import *


description = \
    "Description:\n" + \
    "This tool calculates the significance to the change in mean PSI values between conditions, across multiple conditions.\n" \
    "The conditions are tested in a sequential order specified as input.\n"

parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter,
                        add_help=False)

parser.add_argument('-m', '--method',
                    dest="method",
                    action="store",
                    required=True,
                    choices=['empirical', 'classical'],
                    help="Method to test significance. Required.")

parser.add_argument('-p', '--psi',
                    dest="conds",
                    action="store",
                    nargs="+",
                    help="Path of the PSI files. PSI files and the transcript expression (TPM) files "
                         "must have the same order."
                         "The conditions files and the tpm files must have the same order.")

parser.add_argument('-e', '--tpm',
                    dest="tpms",
                    action="store",
                    nargs="+",
                    help="Path of the transcript expression (TPM) files. Conditions files and the transcript expression "
                         "(TPM) files must have the same order."
                         "The conditions files and the tpm files must have the same order.")

parser.add_argument('-i', '--input',
                    dest="iox",
                    action="store",
                    nargs=1,
                    default=None,
                    help="Input file with the event-transcripts equivalence (.ioe or .ioi format)")

parser.add_argument('-a', '--area',
                    dest="area",
                    action="store",
                    nargs=1,
                    type=int,
                    default=[1000],
                    help="Number indicating the number of points in the local area distribution. (default: 1000)")

parser.add_argument('-l', '--lower-bound',
                    dest="lower_bound",
                    action="store",
                    nargs=1,
                    type=float,
                    default=[0],
                    help="Lower-bound for the absolute delta PSI value to test for significance. (Default: 0).")

parser.add_argument('-pa', '--paired',
                    dest="paired",
                    action="store_true",
                    default=False,
                    help="Boolean. Indicates if replicates in conditions are paired. (Default: False).")

parser.add_argument('-gc', '--gene-correction',
                    dest="gene_cor",
                    action="store_true",
                    default=False,
                    help="Boolean. If True, SUPPA correct the p-values by gene. (Default: False).")

parser.add_argument('-al', '--alpha',
                    dest="alpha",
                    action="store",
                    nargs=1,
                    type=float,
                    default=[0.05],
                    help="Family-wise error rate to use for the multiple test correction. (Default: 0.05).")

parser.add_argument('-s', '--save_tpm_events',
                    action="store_true",
                    default=False,
                    help="Boolean. If True, the average log TPM of the events will be saved in an external file (Default: False).")

parser.add_argument('-c', '--combination',
                    action="store_true",
                    dest="seq",
                    default=False,
                    help="Boolean. If True, SUPPA perform the analysis between all the possible combinations of conditions (Default: False).")

parser.add_argument('-me', '--median',
                    dest="median",
                    action="store_true",
                    default=False,
                    help="Boolean. If True, SUPPA use the median to calculate the Delta PSI. (Default: False).")

parser.add_argument('-th', '--tpm-threshold',
                    dest="tpm_th",
                    action="store",
                    nargs=1,
                    type=float,
                    default=[0.0],
                    help="Minimum transcript average TPM value within-replicates and between-conditions to be included in the analysis. (Default: 1.0).")

def nan_threshold_type(x):
    x = float(x)
    if x < 0.0 or x > 1.0:
        raise ArgumentTypeError("nan_threshold should be a float number between 0 and 1")
    return x

parser.add_argument('-nan', '--nan-threshold',
                    dest="nan_th",
                    action="store",
                    nargs=1,
                    type=nan_threshold_type,
                    default=[0.0],
                    help="Percentage allowed of samples per condition with nan values for returning a DeltaPSI (Default: 0, no missing values allowed).")

parser.add_argument('-o', '--output',
                    dest="output",
                    action="store",
                    default=None,
                    help="Name of the output files.")

parser.add_argument("-mo", "--mode", default="INFO",
                    help="to choose from DEBUG, INFO, WARNING, ERROR and CRITICAL")


def create_path(lst):

    temp_lst = []
    for fl in lst:
        if not os.path.isabs(fl):
            fl_path = os.getcwd()+"/"+fl
            temp_lst.append(fl_path)
        else:
            temp_lst.append(fl)

    return temp_lst


def main():

    args = parser.parse_args()

    # Parsing arguments
    mode = "logging." + args.mode

    # Setting logging preferences
    logger = logging.getLogger(__name__)
    logger.setLevel(eval(mode))

    # Setting the level of the loggers in lib
    # setToolsLoggerLevel(mode)

    # Check if path is absolute, if not the program use the current working path
    cond_files = create_path(args.conds)
    expr_files = create_path(args.tpms)
    ioe_fl = create_path(args.iox)

    # Check extension of input file
    id_type = ioe_fl[0].split(".")[-1].strip("\n")
    if id_type != "ioe" and id_type != "ioi":
        logger.info("Invalid input file. Input file has to be either IOE or IOI format "
                    "it must present the appropriate suffix.")
        exit(1)

    #multiple_conditions_analysis(args.method, cond_files, expr_files, ioe_fl[0], args.area[0], args.lower_bound[0],
     #                            args.paired, args.gene_cor, args.alpha[0], args.output)

    multiple_conditions_analysis(args.method, cond_files, expr_files, ioe_fl[0], args.area[0],
                                 args.lower_bound[0], args.paired, args.gene_cor, args.alpha[0],
                                 args.save_tpm_events, args.seq, args.median, args.tpm_th[0],
                                 args.nan_th[0],args.output)


if __name__ == "__main__":
    main()



# Copyright (c) 2012, Ryan Gomba
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are those
# of the authors and should not be interpreted as representing official policies,
# either expressed or implied, of the FreeBSD Project.

# Modified by Juan L. Trincado on 2017
# Adapted method for using n-dimensions

import math
import json
import numpy as np
import sys


################################################################################
# POINT
################################################################################
from functools import reduce


class Point:
    # def __init__(self, latitude, longitude):
    #     '''Default constructor'''
    #     self.latitude = latitude
    #     self.longitude = longitude
    #     self.cd = None  # core distance
    #     self.rd = None  # reachability distance
    #     self.processed = False  # has this point been processed?

    def __init__(self, id, dimensions):
        '''Constructor for n dimensions'''
        self.id = id # id of the point (e.g: event_name)
        self.dimensions = dimensions # value of points (n dimensions)
        self.cd = None  # core distance
        self.rd = None  # reachability distance
        self.processed = False  # has this point been processed?

    # --------------------------------------------------------------------------
    # calculate the distance between any two points on earth
    # --------------------------------------------------------------------------

    # def distance(self, point):
    #     '''Default distance method for 2 dimensions'''
    #     # convert coordinates to radians
    #
    #     p1_lat, p1_lon, p2_lat, p2_lon = [math.radians(c) for c in
    #                                       (self.latitude, self.longitude, point.latitude, point.longitude)]
    #
    #     numerator = math.sqrt(
    #         math.pow(math.cos(p2_lat) * math.sin(p2_lon - p1_lon), 2) +
    #         math.pow(
    #             math.cos(p1_lat) * math.sin(p2_lat) -
    #             math.sin(p1_lat) * math.cos(p2_lat) *
    #             math.cos(p2_lon - p1_lon), 2))
    #
    #     denominator = (
    #         math.sin(p1_lat) * math.sin(p2_lat) +
    #         math.cos(p1_lat) * math.cos(p2_lat) *
    #         math.cos(p2_lon - p1_lon))
    #
    #     # convert distance from radians to meters
    #     # note: earth's radius ~ 6372800 meters
    #
    #     return math.atan2(numerator, denominator) * 6372800

    def distance(self, point):
        '''Distance method for n dimensions'''

        a = np.array(self.dimensions)
        b = np.array(point.dimensions)
        dist = np.linalg.norm(a - b)

        return dist

    # --------------------------------------------------------------------------
    # point as GeoJSON
    # --------------------------------------------------------------------------

    def to_geo_json_dict(self, properties=None):
        return {
            'type': 'Feature',
            'geometry': {
                'type': 'Point',
                'coordinates': [
                    self.longitude,
                    self.latitude,
                ]
            },
            'properties': properties,
        }

    def __repr__(self):
        return '(%f, %f)' % (self.latitude, self.longitude)


################################################################################
# CLUSTER
################################################################################

class Cluster:
    def __init__(self, points):
        self.points = points

    # --------------------------------------------------------------------------
    # calculate the centroid for the cluster
    # --------------------------------------------------------------------------

    def centroid(self):
        return Point(sum([p.latitude for p in self.points]) / len(self.points),
                     sum([p.longitude for p in self.points]) / len(self.points))

    # --------------------------------------------------------------------------
    # calculate the region (centroid, bounding radius) for the cluster
    # --------------------------------------------------------------------------

    def region(self):
        centroid = self.centroid()
        radius = reduce(lambda r, p: max(r, p.distance(centroid)), self.points)
        return centroid, radius

    # --------------------------------------------------------------------------
    # cluster as GeoJSON
    # --------------------------------------------------------------------------

    def to_geo_json_dict(self, user_properties=None):
        center, radius = self.region()
        properties = {'radius': radius}
        if user_properties: properties.update(user_properties)

        return {
            'type': 'Feature',
            'geometry': {
                'type': 'Point',
                'coordinates': [
                    center.longitude,
                    center.latitude,
                ]
            },
            'properties': properties,
        }


################################################################################
# OPTICS
################################################################################

class Optics:

    def __init__(self, points, max_radius, min_cluster_size):

        self.points = points
        self.max_radius = max_radius  # maximum radius to consider
        self.min_cluster_size = min_cluster_size  # minimum points in cluster

    # --------------------------------------------------------------------------
    # get ready for a clustering run
    # --------------------------------------------------------------------------

    def _setup(self):

        for p in self.points:
            p.rd = None
            p.processed = False
        self.unprocessed = [p for p in self.points]
        self.ordered = []

    # --------------------------------------------------------------------------
    # distance from a point to its nth neighbor (n = min_cluser_size)
    # --------------------------------------------------------------------------

    def _core_distance(self, point, neighbors):

        if point.cd is not None: return point.cd
        if len(neighbors) >= self.min_cluster_size - 1:
            sorted_neighbors = sorted([n.distance(point) for n in neighbors])
            point.cd = sorted_neighbors[self.min_cluster_size - 2]
            return point.cd

    # --------------------------------------------------------------------------
    # neighbors for a point within max_radius
    # --------------------------------------------------------------------------

    def _neighbors(self, point):

        return [p for p in self.points if p is not point and
                p.distance(point) <= self.max_radius]

    # --------------------------------------------------------------------------
    # mark a point as processed
    # --------------------------------------------------------------------------

    def _processed(self, point):

        point.processed = True
        self.unprocessed.remove(point)
        self.ordered.append(point)


    # --------------------------------------------------------------------------
    # update seeds if a smaller reachability distance is found
    # --------------------------------------------------------------------------

    def _update(self, neighbors, point, seeds):

        # for each of point's unprocessed neighbors n...

        for n in [n for n in neighbors if not n.processed]:

            # find new reachability distance new_rd
            # if rd is null, keep new_rd and add n to the seed list
            # otherwise if new_rd < old rd, update rd

            new_rd = max(point.cd, point.distance(n))
            if n.rd is None:
                n.rd = new_rd
                seeds.append(n)
            elif new_rd < n.rd:
                n.rd = new_rd

    # --------------------------------------------------------------------------
    # run the OPTICS algorithm
    # --------------------------------------------------------------------------

    def run(self):

        self._setup()

        # for each unprocessed point (p)...

        cont = 0

        while self.unprocessed:

            point = self.unprocessed[0]
            cont = cont + 1
            percentage = str(round(cont * 100 / len(self.points), 2)) + "%"
            sys.stdout.write("Running OPTICS clustering...%s\r" % percentage)
            sys.stdout.flush()

            # mark p as processed
            # find p's neighbors

            self._processed(point)
            point_neighbors = self._neighbors(point)

            # if p has a core_distance, i.e has min_cluster_size - 1 neighbors

            if self._core_distance(point, point_neighbors) is not None:

                # update reachability_distance for each unprocessed neighbor

                seeds = []
                self._update(point_neighbors, point, seeds)

                # as long as we have unprocessed neighbors...

                while (seeds):

                    # find the neighbor n with smallest reachability distance

                    seeds.sort(key=lambda n: n.rd)
                    n = seeds.pop(0)

                    # mark n as processed
                    # find n's neighbors

                    self._processed(n)
                    n_neighbors = self._neighbors(n)

                    # if p has a core_distance...

                    if self._core_distance(n, n_neighbors) is not None:
                        # update reachability_distance for each of n's neighbors

                        self._update(n_neighbors, n, seeds)

        # when all points have been processed
        # return the ordered list
        percentage = "100%             "
        sys.stdout.write("Running OPTICS clustering...%s\n" % percentage)
        sys.stdout.flush()

        return self.ordered

    # --------------------------------------------------------------------------

    def cluster(self, cluster_threshold):

        clusters = []
        separators = []

        for i in range(len(self.ordered)):
            this_i = i
            next_i = i + 1
            this_p = self.ordered[i]
            this_rd = this_p.rd if this_p.rd else float('infinity')

            # use an upper limit to separate the clusters

            if this_rd > cluster_threshold:
                separators.append(this_i)

        separators.append(len(self.ordered))

        for i in range(len(separators) - 1):
            start = separators[i]
            end = separators[i + 1]
            if end - start >= self.min_cluster_size:
                clusters.append(Cluster(self.ordered[start:end]))

        return clusters


import click


class CustomMultiCommand(click.Group):
    """https://stackoverflow.com/questions/46641928/python-click-multiple-command-names/53144555"""
    ALIASES  = {}
    def command(self, *args, **kwargs):
        """Behaves the same as `click.Group.command()` except if passed
        a list of names, all after the first will be aliases for the first.
        """
        
        def decorator(f):
            if args and args[0]:
                _args = [args[0][0]] + list(args[1:])
                for alias in args[0][1:]:
                    self.ALIASES[args[0][0]] = args[0][1]
                    cmd = super(CustomMultiCommand, self).command(
                        alias, *args[1:], **kwargs)(f)
                    cmd.short_help = "Alias for '{}'".format(_args[0])
            else:
                _args = args
            from click.decorators import command
            cmd = command(*_args, **kwargs)(f)
            self.add_command(cmd)
            return cmd
        return decorator

    def resolve_command(self, ctx, args):
        from click.utils import make_str
        from click.parser import split_opt
        cmd_name = make_str(args[0])
        original_cmd_name = cmd_name

        cmd_name = self.ALIASES.get(cmd_name, cmd_name)
        # Get the command
        cmd = self.get_command(ctx, cmd_name)

        # If we can't find the command but there is a normalization
        # function available, we try with that one.
        if cmd is None and ctx.token_normalize_func is not None:
            cmd_name = ctx.token_normalize_func(cmd_name)
            cmd = self.get_command(ctx, cmd_name)

        # If we don't find the command we want to show an error message
        # to the user that it was not provided.  However, there is
        # something else we should do: if the first argument looks like
        # an option we want to kick off parsing again for arguments to
        # resolve things like --help which now should go to the main
        # place.
        if cmd is None and not ctx.resilient_parsing:
            if split_opt(cmd_name)[0]:
                self.parse_args(ctx, ctx.args)
            ctx.fail(("No such command {name!r}.").format(name=original_cmd_name))
        return cmd_name if cmd else None, cmd, args[1:]


class OptionEatAll(click.Option):
    """Thanks https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click"""

    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):

        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval



"""
Functions and classes for event creation and outputting.

@author: Miha Skalic
@email: miha.skalic[at]gmail.com
"""
import os


class MGene:
    """
    Class holding gene information. For easy event construction
    """
    def __init__(self, gene, chrn, strand):
        self.name = gene.name
        self.chr = chrn
        self.metadata = gene
        self.strand = strand
        self.sortedTranscripts = []
        self.sortedExons = {}
        self.construct_transcripts()
        self.construct_exons()

    def construct_transcripts(self):
        """
        Adding ordered Transcripts to gene
        """
        if len(self.metadata.sortedTranscripts) < 2:
            return
        self.sortedTranscripts = self.metadata.sortedTranscripts

    def construct_exons(self):
        """
        Adding ordered Exons to transcript (gene)
        """
        for transcript in self.sortedTranscripts:
            self.sortedExons[transcript] = sorted(
                self.metadata.transcripts[transcript].exons)


class Event:
    file = ''

    def __init__(self, gene):
        self.gene = gene
        self.alt_search = {}
        self.positive_ids = {}
        self.negative_ids = {}
        self.etype = ''
        self.gtf_string = ('{}\t{}\texon\t{{}}\t{{}}\t.\t{{}}\t.\tgene_id "{{}}"; transcript_id "{{}}:{{}}";\n'
                           .format(self.gene.chr, Event.file))

    def construct_events(self, gene):
        for transcript in gene.sortedTranscripts:
            self.add_putative_events(gene.sortedExons[transcript], transcript)
        for transcript in gene.sortedTranscripts:
            self.add_real_events(gene.sortedExons[transcript], transcript)

    def export_events_ioe(self):
        """
        Returns a list of possible events
        """
        for event in self.positive_ids:
            pos_trans = ','.join(self.positive_ids[event])
            all_trans = ','.join(list(set(self.positive_ids[event] + self.negative_ids[event])))
            full_event = '{};{}:{}'.format(self.gene.name, self.etype, event)

            yield ('{}\t{}\t{}\t{}\t{}\n'.format(self.gene.chr, self.gene.name, full_event,
                                                 pos_trans, all_trans),
                   self.etype)

    def add_putative_events(self, *_):
        pass

    def add_real_events(self, *_):
        pass


class ARSS(Event):
    """
    Alternative 'Right' Splice Site
    """
    def __init__(self, gene, *_):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'A3'
        else:
            self.etype = 'A5'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 2:
            return
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            search_coords = (firstexon[1], secondexon[0])
            self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            self.create_arss(firstexon[0], firstexon[1], secondexon[0], transcript)

    def create_arss(self, coord1, coord2, coord3, transcript):
        gene = self.gene
        for neg1, neg2 in self.negative_coords:
            if neg2 == coord3 and (coord1 < neg1 < coord2):
                eventid = '{}:{}-{}:{}-{}:{}'.format(gene.chr, coord2, coord3, neg1, neg2, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[(neg1, neg2)]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[4]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[3], int(e_vals[3]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line2, self.etype

            line3 = self.gtf_string.format(int(e_vals[4]) - edge, e_vals[4], strand, full_event, full_event,
                                           'alternative2')
            yield line3, self.etype

            line4 = self.gtf_string.format(e_vals[5], int(e_vals[5]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line4, self.etype


class ALSS(Event):
    """
    Alternative 'Left' Splice Site
    """
    def __init__(self, gene, *_):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'A5'
        else:
            self.etype = 'A3'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 2:
            return
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            search_coords = (firstexon[1], secondexon[0])
            self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            self.create_alss(firstexon[1], secondexon[0], secondexon[1], transcript)

    def create_alss(self, coord1, coord2, coord3, transcript):
        gene = self.gene
        for neg1, neg2 in self.negative_coords:
            if neg1 == coord1 and (coord2 < neg2 < coord3):
                eventid = '{}:{}-{}:{}-{}:{}'.format(gene.chr, coord1, coord2, neg1, neg2, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[(neg1, neg2)]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[3], int(e_vals[5]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line2, self.etype

            line3 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line3, self.etype

            line4 = self.gtf_string.format(e_vals[5], int(e_vals[5]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line4, self.etype


class ALTL(Event):
    """
    Alternative on 'left' side of the gene
    """
    def __init__(self, gene, *_):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'AL'
        else:
            self.etype = 'AF'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[0]
        secondexon = exons[1]
        search_coords = (firstexon[0], firstexon[1], secondexon[0])
        self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[0]
        secondexon = exons[1]
        self.create_altl(firstexon[0], firstexon[1], secondexon[0], transcript)

    def create_altl(self, coord1, coord2, coord3, transcript):
        """
        Find and create alternative events
        """
        gene = self.gene
        for neg1, neg2, neg3 in self.negative_coords:
            if neg3 == coord3 and coord2 < neg1:
                eventid = '{}:{}:{}-{}:{}:{}-{}:{}'.format(gene.chr, coord1, coord2, coord3,
                                                           neg1, neg2, neg3, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[(neg1, neg2, neg3)]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(e_vals[2], e_vals[3], strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[4], int(e_vals[4]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line2, self.etype

            line3 = self.gtf_string.format(e_vals[5], e_vals[6], strand, full_event, full_event,
                                           'alternative2')
            yield line3, self.etype

            line4 = self.gtf_string.format(int(e_vals[7]), int(e_vals[7]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line4, self.etype


class ALTR(Event):
    """
    Alternative on 'Right' side of the gene
    """
    def __init__(self, gene, *_):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'AF'
        else:
            self.etype = 'AL'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[-2]
        secondexon = exons[-1]
        search_coords = (firstexon[1], secondexon[0], secondexon[1])
        self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[-2]
        secondexon = exons[-1]
        self.create_altr(firstexon[1], secondexon[0], secondexon[1], transcript)

    def create_altr(self, coord1, coord2, coord3, transcript):
        """
        Find and create alternative events
        """
        gene = self.gene
        for neg1, neg2, neg3 in self.negative_coords:
            if neg1 == coord1 and neg2 > coord3:
                eventid = '{}:{}-{}:{}:{}-{}:{}:{}'.format(gene.chr, coord1, coord2, coord3,
                                                           neg1, neg2, neg3, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[(neg1, neg2, neg3)]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[3], e_vals[4], strand, full_event, full_event,
                                           'alternative2')
            yield line2, self.etype

            line3 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line3, self.etype

            line4 = self.gtf_string.format(e_vals[6], e_vals[7], strand, full_event, full_event,
                                           'alternative1')
            yield line4, self.etype


class MXE(Event):
    """
    Mutually exclusive exons event
    """
    def __init__(self, gene, *_):
        Event.__init__(self, gene)
        self.etype = 'MX'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 3:
            return
        for i in range(len(exons) - 2):
            firstexon = exons[i]
            midexon = exons[i + 1]
            lastexon = exons[i + 2]
            search_coords = firstexon[1], midexon[0], midexon[1], lastexon[0]
            self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 3:
            return
        for i in range((len(exons) - 2)):
            firstexon = exons[i]
            midexon = exons[i + 1]
            lastexon = exons[i + 2]
            search_coords = firstexon[1], midexon[0], midexon[1], lastexon[0]

            for eventid, neg_coord in self.coord_match(search_coords):
                self.negative_ids[eventid] = self.negative_ids.get(eventid, []) + [transcript]
                self.positive_ids[eventid] = self.negative_coords[neg_coord]

    def coord_match(self, new_coords):
        """
        Creates and returns event IDs that match tupled with
        alternative IDs
        """
        all_events = []
        gene = self.gene

        for orig_cord in self.negative_coords:
            compare_f = orig_cord[0] + orig_cord[-1]
            compare_t = new_coords[0] + new_coords[-1]
            mm_f = orig_cord[2]
            mm_t = new_coords[1]

            if compare_f == compare_t and mm_f < mm_t:
                eventid = ('{}:{}-{}:{}-{}:{}-{}:{}-{}:{}'
                           .format(gene.chr, orig_cord[0], orig_cord[1], orig_cord[2], orig_cord[3],
                                   new_coords[0], new_coords[1], new_coords[2], new_coords[3], gene.strand))
                all_events.append([eventid, orig_cord])
        return all_events

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[3], e_vals[4], strand, full_event, full_event,
                                           'alternative1')
            yield line2, self.etype

            line3 = self.gtf_string.format(e_vals[9], int(e_vals[9]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line3, self.etype

            line4 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line4, self.etype

            line5 = self.gtf_string.format(e_vals[7], e_vals[8], strand, full_event, full_event,
                                           'alternative2')
            yield line5, self.etype

            line6 = self.gtf_string.format(e_vals[9], int(e_vals[9]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line6, self.etype


class RI(Event):
    """
    Retained intro event
    """
    def __init__(self, gene, *_):
        Event.__init__(self, gene)
        self.etype = 'RI'

        self.construct_events(gene)

    def add_putative_events(self, exons, transcript):
        gene = self.gene
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            eventid = '{}:{}:{}-{}:{}:{}'.format(gene.chr, firstexon[0], firstexon[1], secondexon[0],
                                                 secondexon[1], gene.strand)

            self.negative_ids[eventid] = self.negative_ids.get(eventid, []) + [transcript]
            search_coord = (firstexon[0], secondexon[1])
            search = self.alt_search.get(search_coord, set())
            search.add(eventid)
            self.alt_search[search_coord] = search

    def add_real_events(self, exons, transcript):
        for i in range(len(exons)):
            search_coord = exons[i]

            if search_coord in self.alt_search:
                eventids = self.alt_search[search_coord]
                for eventid in eventids:
                    self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]

    def export_events_gtf(self, *_):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(e_vals[2], e_vals[3], strand, full_event, full_event, 'alternative2')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[4], e_vals[5], strand, full_event, full_event, 'alternative2')
            yield line2, self.etype

            line3 = self.gtf_string.format(e_vals[2], e_vals[5], strand, full_event, full_event, 'alternative1')
            yield line3, self.etype


class SE(Event):
    """
    Skipped exon event
    """
    def __init__(self, gene, *_):
        Event.__init__(self, gene)
        self.etype = 'SE'

        self.construct_events(gene)
        self.clear_redundant()

    def add_putative_events(self, exons, transcript):
        if len(exons) < 3:
            return
        gene = self.gene
        for i in range(len(exons) - 2):
            firstexon = exons[i]
            midexon = exons[i + 1]
            lastexon = exons[i + 2]
            eventid = '{}:{}-{}:{}-{}:{}'.format(gene.chr, firstexon[1], midexon[0],
                                                 midexon[1], lastexon[0], gene.strand)

            self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
            search_coord = (firstexon[1], lastexon[0])
            search = self.alt_search.get(search_coord, set())
            search.add(eventid)
            self.alt_search[search_coord] = search

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        for i in range((len(exons)-1)):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            search_coord = (firstexon[1], secondexon[0])

            if search_coord in self.alt_search:
                eventids = self.alt_search[search_coord]
                for eventid in eventids:
                    self.negative_ids[eventid] = self.negative_ids.get(eventid, []) + [transcript]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[5], int(e_vals[5]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line2, self.etype

            line3 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line3, self.etype

            line4 = self.gtf_string.format(e_vals[3], e_vals[4], strand, full_event, full_event, 'alternative1')
            yield line4, self.etype

            line5 = self.gtf_string.format(e_vals[5], int(e_vals[5]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line5, self.etype

    def clear_redundant(self):
        for id_key in list(self.positive_ids.keys()):
            if id_key not in self.negative_ids:
                self.positive_ids.pop(id_key)


class EWriter:
    def __init__(self, all_events, output_name, etype, boundary):
        """
        Creates 'write to' buckets
        """
        self.wdict = {}
        for event in all_events:
            if event == 'SS':
                self.wdict['A3'] = open('{}_A3_{}.{}'.format(output_name, boundary, etype), 'w')
                self.wdict['A5'] = open('{}_A5_{}.{}'.format(output_name, boundary, etype), 'w')
            elif event == 'FL':
                self.wdict['AF'] = open('{}_AF_{}.{}'.format(output_name, boundary, etype), 'w')
                self.wdict['AL'] = open('{}_AL_{}.{}'.format(output_name, boundary, etype), 'w')
            else:
                self.wdict[event] = open('{}_{}_{}.{}'.format(output_name, event, boundary, etype), 'w')
        for bucket in self.wdict:
            if etype == 'ioe':
                self.wdict[bucket].write("seqname\tgene_id\tevent_id\talternative_transcripts\ttotal_transcripts\n")
            if etype == 'gtf':
                self.wdict[bucket].write("track name={} visibility=2\n".format(bucket))

    def write(self, line, bucket):
        """
        Given writes the line to the given bucket
        """
        self.wdict[bucket].write(line)

    def close(self):
        """
        Closes all files in writer
        """
        for idx in self.wdict:
            self.wdict[idx].close()


def process_events(my_gene, event, ioe_writer, gtf_writer, edge_len, th):
    """
    Generates specified event occurrences in gene and writes the GTF/IOE output
    """
    gene_event = event(my_gene, th)
    for event_repr, etype in gene_event.export_events_ioe():
        ioe_writer.write(event_repr, etype)
    for gtf_line, etype in gene_event.export_events_gtf(edge_len):
        gtf_writer.write(gtf_line, etype)

# to avoid circular dependencies
from .var_event import *


def create_event_classes(all_events, b_type):
    """
    Creates list of classes that will be calculated for
    """
    if b_type == 'S':
        options = {'SE': SE, 'RI': RI, 'MX': MXE}
    else:
        options = {'SE': SEv, 'RI': RIv, 'MX': MXEv}
    all_classes = []
    for event in all_events:
        if event == 'SS':
            all_classes.append(ARSS if b_type == 'S' else ARSSv)
            all_classes.append(ALSS if b_type == 'S' else ALSSv)
        elif event == 'FL':
            all_classes.append(ALTL if b_type == 'S' else ALTLv)
            all_classes.append(ALTR if b_type == 'S' else ALTRv)
        else:
            all_classes.append(options[event])
    return all_classes


def make_events(events, my_genome, input_name, output_name, edge_len, logger, b_type, th):
    """
    Controls event creation and writing for each gene
    """
    boundary = 'variable_{}'.format(th) if b_type == 'V' else 'strict'
    Event.file = os.path.basename(input_name)
    my_ioe_writer = EWriter(events, output_name, 'ioe', boundary)
    my_gtf_writer = EWriter(events, output_name, 'gtf', boundary)
    my_events = create_event_classes(events, b_type)
    logger.info("Calculating events")
    for gene, xchr, strand in my_genome:
        my_gene = MGene(gene, xchr, strand)
        logger.debug("Analyzing gene: {}".format(gene.name))
        for event in my_events:
            process_events(my_gene, event, my_ioe_writer, my_gtf_writer, edge_len, th)
    my_ioe_writer.close()
    my_gtf_writer.close()



# -*- coding: utf-8 -*-
"""
Created on Wed May 25 04:20:00 CEST 2016

@authors: Juan C Entizne
@email: juancarlos.entizne01[at]estudiant.upf.edu
"""

import os
import argparse
import pandas as pd
from argparse import ArgumentParser, RawTextHelpFormatter


description = \
    "Description:\n" + \
    "This tool joins multiple psi or expression files together.\n" \
    "This tool assume that the first field (column) of the files are in common.\n"

parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter,
                        add_help=False)

parser.add_argument('-i', '--input-files',
                    dest="input",
                    action="store",
                    nargs="+",
                    help="Space separated list of the files to be joined. "
                         "If the absolute path is not indicate the program use the current working directory instead.")

parser.add_argument('-f', '--file-extension',
                    dest="ext",
                    action="store",
                    required=True,
                    choices=['psi', 'tpm'],
                    help="Extension of the output file. Required.")

parser.add_argument('-o', '--output',
                    dest="output",
                    action="store",
                    default=None,
                    help="Name of the output file.")


def create_path(lst):

    temp_lst = []
    for fl in lst:
        if not os.path.isabs(fl):
            fl_path = os.getcwd()+"/"+fl
            temp_lst.append(fl_path)
        else:
            temp_lst.append(fl)

    return temp_lst


def merge_files(fl_lst, output, ext):


    df_lst = []
    for fl in fl_lst:
        df = pd.read_table(fl, sep='\t', index_col=0, header=0)

        old_header = df.columns.values
        new_header = [os.path.basename(fl).split(".")[0]+"_"+col_id for col_id in old_header]
        df.rename(columns=dict(zip(old_header, new_header)), inplace=True)

        df_lst.append(df)

    merged_dfs = pd.concat(df_lst, axis=1)

    header = merged_dfs.columns.values

    with open("%s.%s" % (output, ext), "w+") as fh:
            ln = "\t".join(header)
            fh.write(ln+"\n")

    with open("%s.%s" % (output, ext), "a") as fh:
            merged_dfs.to_csv(fh, sep="\t", na_rep="nan", header=False)


def main():

    args = parser.parse_args()

    input_files = args.input
    outname = args.output
    fl_ext = args.ext

    fl_lst = create_path(input_files)

    merge_files(fl_lst, outname, fl_ext)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Tue May  6 12:19:45 2014

@author: Gael P Alamancos
@email: gael.perez[at]upf.edu
"""

import sys
import logging
import numpy as np
from argparse import ArgumentParser, RawTextHelpFormatter
from lib.tools import *


description = \
    "Description:\n\n" + \
    "This tool reads an ioe file and a transcript expression file and calculates\n" + \
    "the Percentage of Sliced In (PSI)\n"
    
parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter,
                        add_help=False)
parser.add_argument("-i", "--ioe-file", help="Input file with the event-transcripts equivalence (.ioe format).", required=True)
parser.add_argument("-e", "--expression-file", required=True,
                    help="Input transcript expression file.")
parser.add_argument("-o", "--output-file", required=True,
                    help="Output psi file.")
parser.add_argument("-f", "--total-filter", type=float, default=0,
                    help="Minimum total expression of the transcripts involved in the event.")
parser.add_argument('-s', '--save_tpm_events',
                    action="store_true",
                    default=False,
                    help="Boolean. If True, save the TPM of the events in an external file (Default: False).")
parser.add_argument("-m", "--mode", default="INFO",
                    help="to choose from DEBUG, INFO, WARNING, ERROR and CRITICAL")


def main():

    args = parser.parse_args()

    #Parsing arguments
    mode = "logging." + args.mode

    #Setting logging preferences
    logger = logging.getLogger(__name__)
    logger.setLevel(eval(mode))

    # Setting the level of the loggers in lib
    setToolsLoggerLevel(mode)

    output_file = args.output_file
    total_filter = args.total_filter
    expression_file = args.expression_file

    expression_dictionary = {}  # to store transcript expression [transcript_id] = {[colId] = expression}
    psi_dictionary = {}  # to store the psi values calculated [event_id] = {[colId] = expression}
    expression_dictionary_events = {}  # to store the expression values calculated [event_id] = {[colId] = expression}
    col_ids = []  # to store all the column_id for the expression fields
    try:
        #Buffering EXPRESSION FILE
        factory = FactoryReader()
        r = factory.getReader("expression")
        r.openFile(expression_file)
        logger.info("Buffering transcript expression levels.")
        line = r.readLine()
        try:
            while True:
                try:
                    arguments = nextel(line)
                    col_ids = arguments["colIds"]
                    expression_dictionary[arguments["transcript_id"]] = {}
                    # Assign per each colId the expression [colId] = expression
                    for key, value in arguments["expression"].items():
                        expression_dictionary[arguments["transcript_id"]][key] = float(value)
                except ValueError:
                    logger.error("%s expression is not a float. Skipping..." % arguments["transcript_id"])
                    continue
        except StopIteration:
            if not expression_dictionary:
                logger.error("No expression values have been buffered.")
                sys.exit(1)
                
            #Buffering IOE and calculating PSI
            factory = FactoryReader()
            r = factory.getReader("ioe")
            # r.openFile(args.ioe_file)
            r.openFile(args.ioe_file)
            logger.info("Calculating PSI from the ioe file.") 
            line = r.readLine()
            try:
                while True:
                    alternative_transcripts = {}
                    total_transcripts = {}
                    arguments = nextel(line)
                    if psi_dictionary.get(arguments["event_id"]):
                        logger.error("Duplicated event %s. Skipping line..." %
                                     arguments["event_id"])
                        continue
                    skip = False  # to avoid checking total_iso if event == "NA"
                    psi_dictionary[arguments["event_id"]] = {}
                    expression_dictionary_events[arguments["event_id"]] = {}
                    for x in col_ids:
                        #set to 0 all the cumulative expression values
                        alternative_transcripts[x] = 0
                        total_transcripts[x] = 0
                    #Add all the expression values of the alternative transcripts
                    for tr in arguments["alt_iso"].rstrip("\n").split(","):
                        try:
                            for x in col_ids:
                                alternative_transcripts[x] += \
                                    expression_dictionary[tr][x]
                        except KeyError:
                            logger.error(
                                "transcript %s not found in the \"expression file\"." %
                                tr)
                            logger.error("PSI not calculated for event %s." %
                                         arguments["event_id"])
                            for x in col_ids:
                                psi_dictionary[arguments["event_id"]][x] = "NA"
                                expression_dictionary_events[arguments["event_id"]][x] = "NA"
                            skip = True
                    # Add all the expression values of the total transcripts
                    if not skip:
                        for tr in arguments["total_iso"].rstrip("\n").split(","):
                            try:
                                for x in col_ids:
                                    total_transcripts[x] += \
                                        expression_dictionary[tr][x]
                            except KeyError:
                                logger.error(
                                    "transcript %s not found in the \"expression file\"." %
                                    tr)
                                logger.error("PSI not calculated for event %s." %
                                             arguments["event_id"])
                                for x in col_ids:
                                    psi_dictionary[arguments["event_id"]][x] = "NA"
                                    expression_dictionary_events[arguments["event_id"]][x] = "NA"
                                skip = True
                    for x in col_ids:
                        #If it passes the filter and skip == False
                        if total_transcripts[x] >= total_filter and not skip:
                            try:
                                psi_dictionary[arguments["event_id"]][x] = (
                                    alternative_transcripts[x]/total_transcripts[x])
                                expression_dictionary_events[arguments["event_id"]][x] = total_transcripts[x]
                            except ZeroDivisionError:
                                logger.debug("Zero division for event %s.(psi= nan)." % arguments["event_id"])
                                psi_dictionary[arguments["event_id"]][x] = np.nan
                                expression_dictionary_events[arguments["event_id"]][x] = total_transcripts[x]
                        #If it passes the filter but skip == True
                        elif not skip:   
                            #for x in col_ids:
                            psi_dictionary[arguments["event_id"]][x] = np.nan
                            expression_dictionary_events[arguments["event_id"]][x] = total_transcripts[x]

            except StopIteration:
                writer = Writer.getWriter("PSI")
                logger.info("Generating output %s" % (output_file + ".psi"))
                writer.openFile(output_file)
                writer.writeLine("\t".join(col_ids), False)
                for key, value in sorted(psi_dictionary.items()):
                    logger.debug("Calculating psi for %s" % key)
                    psi_line = PsiWriter.lineGenerator(key, value, col_ids)
                    writer.writeLine("\t".join(psi_line))
                writer.closeFile()

                if (args.save_tpm_events == True):
                    #Save the expression of the events
                    writer = Writer.getWriter("TPM")
                    logger.info("Generating output %s" % (output_file + ".tpm"))
                    writer.openFile(output_file)
                    writer.writeLine("\t".join(col_ids), False)
                    for key, value in sorted(expression_dictionary_events.items()):
                        logger.debug("Calculating tpm for %s" % key)
                        tpm_line = TpmWriter.lineGenerator(key, value, col_ids)
                        writer.writeLine("\t".join(tpm_line))
                    writer.closeFile()

    except BaseException:
        logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
        sys.exit(1)
    logger.info("Done")

if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
"""
Created on Wed Aug 06 17:51:05 2014

@author: Gael P Alamancos
@email: gael.perez[at]upf.edu
"""

import fileMerger as joinFiles
import psiPerGene as psiPerIsoform
import psiCalculator as psiPerEvent
import eventGenerator as generateEvents
import eventClusterer as clusterAnalysis
import significanceCalculator as diffSplice
import logging
import argparse 
import sys


description = "Description:\n\n" + \
              "SUPPA allows you to generate all the possible Alternative Splicing events from an annotation file, \n" \
              "calculate the PSI values per event, calculate differential splicing across multiple conditions \n" \
              "with replicates, and cluster events across conditions \n" \
              "For further information, see the help of each subcommand."

parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter, prog='SUPPA')
# Version
parser.add_argument('-v','--version', action='version', version='%(prog)s 2.3')
subparsers = parser.add_subparsers()

# EventGenerator parser
eventGeneratorSubparser = subparsers.add_parser(
    "generateEvents", parents=[generateEvents.parser],
    help="Calculates the Alternative Splicing events from an annotation file.")
eventGeneratorSubparser.set_defaults(which="generateEvents")

# psiCalculator parser
psiCalculatorSubparser = subparsers.add_parser(
    "psiPerEvent", parents=[psiPerEvent.parser],
    help="Calculates the PSI value for each event previously generated.")
psiCalculatorSubparser.set_defaults(which="psiPerEvent")

# psiPerGene parser
psiPerGeneSubparser = subparsers.add_parser(
    "psiPerIsoform", parents=[psiPerIsoform.parser],
    help="Calculates the PSI value for each isoform.")
psiPerGeneSubparser.set_defaults(which="psiPerIsoform")

# significanceCalculator parser
significanceCalculatorSubparser = subparsers.add_parser(
    "diffSplice", parents=[diffSplice.parser],
    help="Calculates differentially spliced events across multiple conditions.")
significanceCalculatorSubparser.set_defaults(which="diffSplice")

# eventClusterer parser
eventClustererSubparser = subparsers.add_parser(
    "clusterEvents", parents=[clusterAnalysis.parser],
    help="Calculates clusters of events across conditions.")
eventClustererSubparser.set_defaults(which="clusterEvents")

# fileMerger parser
fileMergerSubparser = subparsers.add_parser(
    "joinFiles", parents=[joinFiles.parser],
    help="Join multiple tab separated files into a single file.")
fileMergerSubparser.set_defaults(which="joinFiles")

# Setting logging preferences
logger = logging.getLogger(__name__)


def main():

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    try:
        args = parser.parse_args()
        if args.which == "generateEvents":
            generateEvents.parser = parser  # Setting the module aparser
            generateEvents.main()
        elif args.which == "psiPerEvent":
            psiPerEvent.parser = parser  # Setting the module parser
            psiPerEvent.main()
        elif args.which == "psiPerIsoform":
            psiPerIsoform.parser = parser  # Setting the module parser
            psiPerIsoform.main()
        elif args.which == "diffSplice":
            diffSplice.parser = parser  # Setting the module parser
            diffSplice.main()
        elif args.which == "clusterEvents":
            clusterAnalysis.parser = parser  # Setting the module parser
            clusterAnalysis.main()
        elif args.which == "joinFiles":
            joinFiles.parser = parser  # Setting the module parser
            joinFiles.main()
    except Exception:
        logger.error("Unknown error: {}".format(sys.exc_info()))
        sys.exit(1)
        
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 19 16:27:06 2013

@author: Gael P. Alamancos
@email: gael.perez[at]upf.edu
"""

import sys
import logging
from abc import ABCMeta, abstractmethod
from .event import *


#Setting logging preferences
logging.basicConfig(level=logging.INFO)    
logger = logging.getLogger(__name__)


def nextel(my_object):
    """
    Returns next element of generator. Python2 and Python3 compatible.
    """
    if sys.version_info >= (3, 0):
        return my_object.__next__()
    else:
        return my_object.next()


def setToolsLoggerLevel(mode):
    try:
        logger.setLevel(eval(mode))
    except:
        logger.warning("Could not set up tools logger into %s level. Using INFO." %
            mode)


class FormatError(Exception):
    def __init__(self, line = None, *args):
        '''Exception raised for errors in the format of the input file.
        
        Keyword arguments:
        
        line            -- input line number where the error occured
        message (*args) -- accumulated errors explanation, i.e., tuple with the 
        explanations of the errors
        '''
        self.line = line        
        self.message = args  


class Parser(object):
    '''Abstract Class containing the common behaviour of all the parsers'''
    __metaclass__=ABCMeta
    
    def parseLine(self, line, lineNumber):
        '''General line parsing.
        
        Keyword arguments:
        
        line -- the input line
        lineNumber -- the line number of the input file
        
        Observations:
        
        The following attributes should be define in a subclass since it 
        depends on the filetype.
        
        self.MIN_FIELDS, self.SEQ_NAME, self.FEATURE_INDEX, self.START_INDEX, 
        self.END_INDEX, self.STRAND_INDEX, self.ATTR_INDEX 
        
        other attributes may be define if appropriate. 
        
        '''
        try:
            msg = []    #It will containg the error message acummulated.
            fields = line.rstrip('\n').split('\t')
            #Parsing the number of fields
            if len(fields) < self.MIN_FIELDS:
                msg.append("Unexpected number of fields.")
            else:
                #Checking for empty or blank(\s) fields
                for empty in (i for i in range(self.MIN_FIELDS) if not \
                    fields[i] or fields[i].isspace()):                    
                    msg.append("Field %i empty or blank" % (empty+1))
                #Checking format of some fields
                if not msg:
                    try:
                        if int(fields[self.START_INDEX]) > int(
                            fields[self.END_INDEX]):
                            msg.append("Start coordinate greater than end "+ \
                                "coordinate.")
                    except ValueError:
                        msg.append("Start, end or both coordinates are not "+ \
                            "integers.")
                    if fields[self.STRAND_INDEX] not in ('+', '-', '.'):
                        msg.append("Unknown strand format in field: %i" % (
                            self.STRAND_INDEX+1))
            if msg: 
                raise FormatError(lineNumber, msg)
            return True
        except FormatError:
            logger.error("%s, in line %i. %s. Skipping line..." % (sys.exc_info()[1].args[0], lineNumber + 1, msg[0]))
            return False
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)


class IoeParser(Parser):
    def __init__(self):    
        '''It contains the correspondace of field in the IOE file starting 
            with 0.
        '''
        self.MIN_FIELDS = 5 #Number of fields
        self.SEQNAME_INDEX = 0 
        self.GENE_ID_INDEX = 1
        self.EVENT_ID_INDEX = 2
        self.ALT_ISO_INDEX = 3
        self.TOTAL_ISO_INDEX = 4
        
    def parseLine(self, line, lineNumber):
        try:
            msg = []    #It will containg the error message acummulated.
            fields = line.rstrip('\n').split('\t')
            #Parsing the number of fields
            if len(fields) < self.MIN_FIELDS:
                msg.append("Unexpected number of fields.")
            else:
                #Checking for empty or blank(\s) fields
                for empty in (i for i in range(self.MIN_FIELDS) if not \
                    fields[i] or fields[i].isspace()):                    
                    msg.append("Field %i empty or blank" % (empty+1))
                #Checking format of some fields
                if not msg:
                    alt1 = set(fields[self.ALT_ISO_INDEX].split(','))
                    total = set(fields[self.TOTAL_ISO_INDEX].split(','))
                    if not total.issuperset(alt1):
                        msg.append ("Not all alternative_trancripts (field %i) present in the total_trancripts (field %i)" % (
                            self.ALT_ISO_INDEX+1, self.TOTAL_ISO_INDEX+1))
            if msg: 
                raise FormatError(lineNumber, msg)
            return True
        except FormatError:
            logger.error("%s, in line %i. %s. Skipping line..." % (sys.exc_info()[1].args[0], lineNumber + 1, msg[0]))
            return False
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1) 
            
            
class ExpressionParser(Parser):
    def __init__(self):
        self.MIN_FIELDS = 2 #Number of fields
        self.TRANSCRIPT_ID_INDEX = 0
        self.EXPRESSION_INDEX = 1
        
#    def parseLine(self, line, lineNumber, tableFormat = False):
    def parseLine(self, line, lineNumber):
        '''Parse a expression line 
        
        keyword arguments:

        line -- line of an expression file.
        lineNumber -- line number to display in case of format errors.
        tableFormat -- true when there's more than one expression field.        
        '''
        try:
            msg = []    #It will containg the error message acummulated.
            fields = line.rstrip('\n').split('\t')
            #Parsing the number of fields
            if len(fields) < self.MIN_FIELDS:
                msg.append("Unexpected number of fields. %i expected, %i given" % \
                    (self.MIN_FIELDS, len(fields)))
            else:
                #Checking for empty or blank(\s) fields
                for empty in (i for i in range(self.MIN_FIELDS) if not \
                    fields[i] or fields[i].isspace()):                    
                    msg.append("Field %i empty or blank" % (empty+1))
                
                #Checking format of all expression fields
                for i in range(1, self.MIN_FIELDS):
                    try:
                        float(fields[i])
                    except ValueError:
                        msg.append("Field %i is not a float." % (i+1))
                        break
            if msg: 
                raise FormatError(lineNumber, msg)
            return True
        except FormatError:
            logger.error(msg)
            logger.error("%s, in line %i. %s. Skipping line..." % (sys.exc_info()[1].args[0], lineNumber + 1, msg[0]))
            return False
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1) 
            

class PsiParser(Parser):
    def __init__(self):
        self.MIN_FIELDS = 2 #Number of fields
        self.EVENT_ID_INDEX = 0
        self.PSI_INDEX = 1
        
#    def parseLine(self, line, lineNumber, tableFormat = False):
    def parseLine(self, line, lineNumber):
        try:
            msg = []    #It will containg the error message acummulated.
            fields = line.rstrip('\n').split('\t')
            #Parsing the number of fields
            if len(fields) < self.MIN_FIELDS:
                msg.append("Unexpected number of fields.")
            else:
                #Checking for empty or blank(\s) fields
                for empty in (i for i in range(self.MIN_FIELDS) if not \
                    fields[i] or fields[i].isspace()):                    
                    msg.append("Field %i empty or blank" % (empty+1))
                
                #Checking format of all psi fields
                for i in range(1, self.MIN_FIELDS):
                    try:
                        float(fields[i])
                    except ValueError:
                        if fields[i] != "NA":
                            msg.append("Field %i is not a float or \"NA\"." % (i+1))
                            break                
                if msg: 
                    raise FormatError(lineNumber, msg)
            return True
        except FormatError:
            logger.error("%s, in line %i. %s. Skipping line..." % (sys.exc_info()[1].args[0], lineNumber + 1, msg[0]))
            return False
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)


class TpmParser(Parser):
    def __init__(self):
        self.MIN_FIELDS = 2  # Number of fields
        self.EVENT_ID_INDEX = 0
        self.PSI_INDEX = 1

    #    def parseLine(self, line, lineNumber, tableFormat = False):
    def parseLine(self, line, lineNumber):
        try:
            msg = []  # It will containg the error message acummulated.
            fields = line.rstrip('\n').split('\t')
            # Parsing the number of fields
            if len(fields) < self.MIN_FIELDS:
                msg.append("Unexpected number of fields.")
            else:
                # Checking for empty or blank(\s) fields
                for empty in (i for i in range(self.MIN_FIELDS) if not \
                        fields[i] or fields[i].isspace()):
                    msg.append("Field %i empty or blank" % (empty + 1))

                # Checking format of all psi fields
                for i in range(1, self.MIN_FIELDS):
                    try:
                        float(fields[i])
                    except ValueError:
                        if fields[i] != "NA":
                            msg.append("Field %i is not a float or \"NA\"." % (i + 1))
                            break
                if msg:
                    raise FormatError(lineNumber, msg)
            return True
        except FormatError:
            logger.error("%s, in line %i. %s. Skipping line..." % (sys.exc_info()[1].args[0], lineNumber + 1, msg[0]))
            return False
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)


class Reader(object):
    '''Abstract Class with the common behaviour of all the readers.'''
    __metaclass__ = ABCMeta
    
    def openFile(self, fileName):
        '''It creates a pipe for reading the input file'''
        try:
            self._pipe = open(fileName, 'r')
            logger.info("File %s opened in reading mode." % fileName)
        except IOError:
            logger.error("I/O Error: %s" % (sys.exc_info()[1].args[1]))
            sys.exit(1)
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)

    def readLine(self, header = False):
        '''This method requires to have a Reader sublass that inherits also from a 
        Parser class'''
        try:        
            for lineNumber, line in enumerate(self._pipe):
                #Parsing the line. Line skipped if there's any error. 
                #fields[0:8]
                if lineNumber == 0 and header: #Skip the header line
                    continue
                if line.startswith('#'):
                    logger.debug("Line %i starts with #. Skipping line..." % (
                    lineNumber+1))
                    continue
                if not self.parseLine(line, lineNumber):
                    continue
                else:
                    arguments = {}
                    fields = line.rstrip("\n").split("\t")
                    arguments["seqname"] = fields[self.SEQNAME_INDEX]
                    arguments["feature"] = fields[self.FEATURE_INDEX]
                    arguments["start"] = fields[self.START_INDEX]
                    arguments["end"] = fields[self.END_INDEX]
                    arguments["strand"] = fields[self.STRAND_INDEX]
                    attributes = self.parseAttributes(
                        fields[self.ATTR_INDEX], lineNumber,
                        arguments["feature"])
                    #Parsing attributes. Line skipped if there's any error. 
                    #fields[8]                  
                    if attributes == False:
                        continue
                    elif isinstance(attributes, dict):
                        arguments.update(attributes)
                    yield arguments
            self._pipe.close()
            logger.info("File %s closed." % self._pipe.name)
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)

        
class IoeReader(Reader, IoeParser):
    def __init__(self):
        IoeParser.__init__(self)
        
    def readLine(self, header = True):
        '''This method reads and parses an ioe line.'''
        try:        
            for lineNumber, line in enumerate(self._pipe):
                if lineNumber == 0 and header: #Skip the header line
                    continue
                if line.startswith('#'):
                    logger.debug("Line %i starts with #. Skipping line..." % (
                    lineNumber+1))
                    continue
                if not self.parseLine(line, lineNumber):
                    continue
                else:
                    arguments = {}
                    fields = line.rstrip("\n").split("\t")
                    arguments["seqname"] = fields[self.SEQNAME_INDEX]
                    arguments["gene_id"] = fields[self.GENE_ID_INDEX]
                    arguments["event_id"] = fields[self.EVENT_ID_INDEX]
                    arguments["alt_iso"] = fields[self.ALT_ISO_INDEX]
                    arguments["total_iso"] = fields[self.TOTAL_ISO_INDEX]
                    yield arguments
            self._pipe.close()
            logger.info("File %s closed." % self._pipe.name)
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)


class ExpressionReader(Reader, ExpressionParser):
    def __init__(self):
        ExpressionParser.__init__(self)
        
#    def readLine(self, header = True, tableFormat = False):
    def readLine(self, header = True):
        '''This method reads and parses an "expression" line'''
        try:
            min_fields = 0
            colIds = list() 
            for lineNumber, line in enumerate(self._pipe):
                fields = line.rstrip("\n").split("\t")
                if lineNumber == 0 and header: #Skip the header line
                    #Calculating the number of fields required
                    min_fields = (len(line.rstrip("\n").split("\t")) + 1)
                    #Storing column_id for the expression fields
                    colIds = fields
                    continue
                if line.startswith('#'):
                    logger.debug("Line %i starts with #. Skipping line..." % (
                        lineNumber+1))
                    continue
#                if tableFormat and min_fields > 2:
                if min_fields > 2:
                    #update the number of fields to request in each line
                    self.MIN_FIELDS = min_fields    
                #Parse the line
#                if not self.parseLine(line, lineNumber, tableFormat):
                if not self.parseLine(line, lineNumber):
                     continue
                else:
                    arguments = {}
                    arguments["transcript_id"] = fields[self.TRANSCRIPT_ID_INDEX]
                    arguments["expression"] = {}
                    arguments["colIds"] = colIds
                    for x in range(1, self.MIN_FIELDS):
                        arguments["expression"][colIds[(x-1)]] = fields[x]                        
                    yield arguments
            self._pipe.close()
            logger.info("File %s closed." % self._pipe.name)    
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)


class FactoryReader(object):
    def getReader(self, fileFormat):
        if fileFormat == "ioe":
            return IoeReader()
        elif fileFormat == "expression":
            return ExpressionReader()
        else:
            logger.error("Unknown reader:" + fileFormat)
            sys.exit(1)


class Writer(object):
    '''Abstract class that encapsulates all the common behaviour of the 
    writers
    '''
    __metaclass__ = ABCMeta
    
    def openFile(self, outputFile):
        if (not os.path.exists(os.path.dirname(outputFile))) and (os.path.dirname(outputFile) != '') :
            logger.error('Can not save to file "{}". Path does not exist.'.format(outputFile))
            sys.exit(1)
        try:
            self._pipe = open(outputFile, 'w')
            self.lineNumber = 0    
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)

    def writeLine(self, line, parse = True):
        self.write(line+"\n", parse)
        
    def write(self, line, parse = True):
        '''This method requires to have a Writer sublass that inherits also 
        from a Parser class'''
        try:
            if not line.startswith('#') and parse:
                if self.parseLine(line, self.lineNumber):
                    self._pipe.write(line)    
                    self.lineNumber += 1
            else:
                self._pipe.write(line)
                self.lineNumber += 1
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)

    @staticmethod
    def getWriter(fileFormat):
        try:
            if fileFormat == "GTF" or fileFormat == "IOE":
                logger.error("Error: These GTF and IOE writers no longer exist.")
                sys.exit(1)
            elif fileFormat == "PSI":
                return PsiWriter()
            elif fileFormat == "TPM":
                return TpmWriter()
            else:
                logger.error("Unknown writer:" + fileFormat)
                sys.exit(1)
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)

    def closeFile(self):
        try:
            self._pipe.close()
        except BaseException:
            logger.error("Unknown error: %s" % sys.exc_info()[1].args[0])
            sys.exit(1)


class PsiWriter(Writer, PsiParser):
    '''It conatains all the utilities required to write an PSI file.'''
    def openFile(self, outputFile):
        outputFile += ".psi"
        Writer.openFile(self, outputFile)
        
    @staticmethod
    def lineGenerator(key, value, colIds):
        '''Generates a PSI line from a single event
        
        key -- id of the event.
        value -- the dictionary containning all column names and psi values calculated.
        colIds -- list containing the column names
        '''
        line = []
        line.append(key)
        for x in colIds:
            line.append(str(value[x]))
        return line


def ioi_writer(genome_obj, output_file):

    with open(output_file+".ioi", "w+") as fh:
        fh.write("seqname\tgene_id\tisoform_id\tinclusion_transcripts\ttotal_transcripts\n")
        for obj in genome_obj:
            for trans_id in obj[0].transcripts.keys():
                chrm = obj[1]
                gene_id = obj[0].name
                all_trans_ids = ",".join(obj[0].sortedTranscripts)
                line = "%s\t%s\t%s\t%s\t%s\n" % \
                       (chrm, gene_id, gene_id+";"+trans_id, trans_id, all_trans_ids)
                fh.write(line)

                
class TpmWriter(Writer, TpmParser):
    '''It conatains all the utilities required to write an PSI file.'''

    def openFile(self, outputFile):
        outputFile += ".tpm"
        Writer.openFile(self, outputFile)

    @staticmethod
    def lineGenerator(key, value, colIds):
        '''Generates a PSI line from a single event

        key -- id of the event.
        value -- the dictionary containning all column names and psi values calculated.
        colIds -- list containing the column names
        '''
        line = []
        line.append(key)
        for x in colIds:
            line.append(str(value[x]))
        return line




import sys
import json
import hashlib
import logging
from pathlib import Path
from functools import partial
from itertools import chain, repeat

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans

from .suppa.lib.event import make_events
from .suppa.lib.gtf_store import *
from .suppa.lib.tools import *
from . import feature_len as fl


def sig_filter(df, dpsi=0, abs_dpsi=0, pval=0.05):
    dpsi_df = df
    pval_keep = dpsi_df.loc[:, "pval"] < pval
    if dpsi > 0:
        keep = (dpsi_df.loc[:, "dpsi"] > dpsi) & pval_keep
    elif dpsi < 0:
        keep = (dpsi_df.loc[:, "dpsi"] < dpsi) & pval_keep
    else:
        keep = pval_keep
    if abs_dpsi > 0:
        keep = (abs(dpsi_df.loc[:, "dpsi"]) > abs_dpsi) & keep
    fdf = dpsi_df.loc[keep, ]
    return fdf

def compute_feature_len(df):
    
    lens = df["event_id"].apply(fl.AS_len)
    len_df = pd.DataFrame(lens.tolist())
    return len_df

def extract_info(df):
    from . import feature_len as fl
    chrs = df["event_id"].apply(lambda x: x.split(":")[1])
    genes = df["event_id"].apply(lambda x: x.split(";")[0].strip())
    lens = df["event_id"].apply(fl.AS_len)
    len_df = pd.DataFrame(lens.tolist())
    data = {
        "geneId": genes,
        "chr": chrs,
    }
    new_df = pd.DataFrame(data)
    len_df.index = new_df.index
    new_df = pd.concat([new_df, len_df], axis=1)
    return new_df

def check_gtf_used(gtf):
    gtf = Path(gtf)
    gtf_hash = hashlib.blake2b()
    gtf_size = gtf.stat().st_size
    gtf_hash.update(str(gtf_size).encode('utf-8'))
    f = gtf.open("rb")
    for _ in range(10):
        gtf_hash.update(f.readline().strip())
    hash_val = gtf_hash.hexdigest()
    return hash_val

def ioe_event(gtf, event_type, edge_exon_len):

    mode = "logging.INFO" 

    # Setting logging preferences
    logger = logging.getLogger(__name__)
    logger.setLevel(eval(mode))

    # Setting the level of the loggers in lib
    # setToolsLoggerLevel(mode)

    home = Path.home()
    laas_dir = home / f".astk"
    laas_dir.mkdir(exist_ok=True)

    ref_dir = laas_dir / "ref"
    ref_dir.mkdir(exist_ok=True)

    gtf_hash = check_gtf_used(gtf)
    output_dir = ref_dir / gtf_hash
    
    if not output_dir.exists():
        output_dir.mkdir(exist_ok=True)
        my_genome = Genome()
        logger.info("Reading input data.")
        try:
            fetched_exons = gtf_reader(gtf, logger)
        except UnicodeDecodeError as e:
            print(e)
            sys.exit(1)

        if len(fetched_exons) == 0:
            logger.info("No exons found. Check format and content of your GTF file.")
            sys.exit(1)

        for exon_meta in fetched_exons:
            my_genome.add_to_genes(exon_meta)
        
        my_genome.sort_transcripts()
        my_genome.split_genes()
        logger.info("Pooling genes")
        my_genome.poll_genes()

        out_prefix = output_dir / "annotation"
        make_events(event_type, my_genome, gtf, out_prefix, edge_exon_len,
                logger, b_type="S", th=10)
    else:
         logger.info("Loading cache data.")
    return output_dir


def df_generate(path, rep, group, group_name):
    if len(rep) == 1:
        reps = [int(rep[0])] * group
    elif len(rep) == group:
        reps = [int(i) for i in rep]
    else:
        print("rep dismatch group")
        sys.exit()
    if len(path) == 1:
        paths = [path[0] for _ in range(sum(reps))]
    elif len(path) == int(rep[0]):
        paths = list(chain(*repeat(path,  group)))
    elif len(path) == sum(reps):
        paths = path
    else:
        print("path number dismath")
        sys.exit()
    
    rep_ls = list(chain(*((range(1, i+1)) for i in reps)))

    if group_name is None:
        groups = list(chain(*[repeat(i, r) for r,i in zip(reps, range(1, group+1))]))
    elif len(group_name) != group:
        print("group name number is not consistent with group number, laas will using default value")
        groups = list(chain(*[repeat(i, r) for r,i in zip(reps, range(1, group+1))]))
    elif len(group_name) == group:
        
        groups = list(chain(*[repeat(i, r) for r,i in zip(reps, range(1, group+1))]))
        groups = [group_name[i-1] for i in groups]
    
    names = [Path(i).parent.name for i in paths] 
    df = pd.DataFrame({
        "group": groups,                                                                                                                                                                                                                                                                               
        "replicate": rep_ls,
        "name": names,
        "path": paths
    })
    return df


def meta_template(out, group, repN, group_name,
             path1, path2, repN1=None, repN2=None):

    if all([repN1, repN2]):
        control_df1 = df_generate(path1, repN1, group, group_name)
        treatment_df2 = df_generate(path2, repN2, group, group_name)
    elif any([repN1, repN2]):
        print("-repN1 and -repN2 must be set simultaneously")
    elif repN:
        control_df1 = df_generate(path1, repN, group, group_name)
        treatment_df2 = df_generate(path2, repN, group, group_name)
    control_df1.insert(1, "condition", "control")
    treatment_df2.insert(1, "condition", "treatment")
    df = pd.concat([control_df1, treatment_df2]).sort_values(by=["group", "condition"])
    df.to_csv(out, index=False)
    
    meta_dic = {}
    for gn, gdf in df.groupby("group"):
        meta_dic[gn] = {"control": {"samples": []},
                         "treatment": {"samples": []}}
        for _, row in gdf.iterrows():
            name = row["name"]
            rep = row["replicate"]
            path = row["path"]
            sp_dic = {"name": name, "replicate": rep, "path": path}
            meta_dic[gn][row["condition"]]["samples"].append(sp_dic)
    
    outjson = Path(out).with_suffix(".json")

    f = outjson.open(mode="w")
    json.dump(meta_dic, f, indent=4)
    f.close()


def ioe_psi(event_row, tpm):
    alternative_transcripts = event_row["alternative_transcripts"].split(",")
    total_transcripts = event_row["total_transcripts"].split(",")
    al_tpts_val = [i for i in alternative_transcripts if i in tpm.index]
    all_tpts_val = [i for i in total_transcripts if i in tpm.index]

    if len(all_tpts_val) == 0:
        psi_ls = ["nan" for _ in tpm.columns]
    elif len(al_tpts_val) == 0 and len(all_tpts_val) > 0:
        psi_ls = [0 for _ in tpm.columns]
    else:
        get_val = lambda tid, colid: tpm.loc[tid, colid]
        psi_ls = []
        for colid in tpm.columns:
            get_val = lambda tid: tpm.loc[tid, colid]
            al_tpts_abundance = sum(map(get_val, al_tpts_val))
            all_tpts_abundance = sum(map(get_val, all_tpts_val))
            if all_tpts_abundance < 0.001:
                psi_ls.append("nan")
                continue
            try:
                psi = al_tpts_abundance / all_tpts_abundance
            except ZeroDivisionError:
                psi = "nan"
            psi_ls.append(psi)
    return psi_ls


def len_hist(len_counts, width, max_len):
    offset = (width - max_len % width) if max_len % width else 0
    bins =  (max_len + offset) // width
    counts, bin_edges = np.histogram(len_counts, bins=bins, range=(1, max_len+offset+1))
    return counts, bin_edges


def plot_hist_cluster(out, cluster_ls, bins):
    fig, ax = plt.subplots()
    for i, subset in enumerate(cluster_ls, 1):
        ax.hist(subset, bins=bins, alpha=0.5, label=f"Cluster {i}")
    ax.legend()
    plt.savefig(Path(out).with_suffix(".png"))  

                                                         
def custome_cluster_len(df, out, lens, width=10, max_len=500):
    len_count = df["len"]
    counts, bin_edges = len_hist(len_count, width, max_len)
    cluster_ls = []
    start = 0
    for idx, item in enumerate(lens):
        tmp = len_count[start < len_count]
        subset = tmp[tmp <= item]
        cluster_ls.append(subset)

        df.loc[(start < len_count) & (len_count <= item), "cluster"] = idx+1
        start = item
    else:
        cluster_ls.append(len_count[len_count > item])
        df.loc[len_count >= item, "cluster"] = idx + 2

    lps = [1, *lens, max(bin_edges)]
    cn_ls = [f"cluster{idx+1}" for i in range(len(lens)+1)]
    range_ls = [f"{lps[i]}-{lps[i+1]}" for i in range(len(lps)-1)]
    count_ls = [len(i) for i in cluster_ls]
    cluster_info = pd.DataFrame({"cluster": cn_ls, "range": range_ls, "counts": count_ls})
    cluster_info.to_csv(Path(out).with_suffix(".cluster.csv"))

    plot_hist_cluster(out, cluster_ls, bin_edges)
    return df


def cluster_len(df, out, n_cls=5, width=10, max_len=500, len_weight = 5):
    len_count = df["len"]
    counts, bin_edges = len_hist(df["len"], width, max_len)
    lens = [(bin_edges[i] + bin_edges[i+1])/2-1  for i,v in enumerate(bin_edges[:-1])]
    lens = np.array(lens).reshape(-1,1) * len_weight
    counts = counts.reshape(-1,1)
    data = np.concatenate([counts, lens], axis = 1)

    cluster_id = KMeans(n_cls, random_state=2).fit_predict(data)
    new_cls_dic = {}

    range_ls = []
    for ci in np.unique(cluster_id):
        cls_idx = list(np.where(cluster_id == ci)[0])  
        start = bin_edges[cls_idx[0]]
        end = bin_edges[cls_idx[-1] + 1]
        tmp = len_count[start <= len_count]
        subset = tmp[tmp < end]
        new_cls_dic[start] = subset
        range_ls.append([start, end])

    cluster_ls = [new_cls_dic[i] for i in sorted(new_cls_dic.keys())]

    plot_hist_cluster(out, cluster_ls, bin_edges)

    range_ls = sorted(range_ls, key=lambda x: x[0])

    range_str_ls = [f"{i[0]}-{i[1]}" for i in range_ls]
    count_ls = [len(i) for i in cluster_ls]

    for idx, (s, e) in enumerate(range_ls, 1):
        print(s, e)
        df.loc[(s <= df["len"]) & ( df["len"] < e), "cluster"] = idx
        print(df.head())
    else:
         df.loc[df["len"] >= e, "cluster"] = idx

    cn_ls = [f"cluster{i+1}" for i in range(len(cluster_ls))]
    cluster_info = pd.DataFrame({"cluster": cn_ls, "range": range_str_ls, "counts": count_ls})
    cluster_info.to_csv(Path(out).with_suffix(".cls.csv") )
    print(df.head())
    return df


class DiffSplice:
    def __init__(self, outdir, meta, gtf, as_types,
                 exon_len) -> None:
        self.psi = {}
        self.mk_outdir(outdir)
        self.ioe_dir = self.get_ioe_event(gtf, exon_len)
        self.parse_meta(meta)
        for at in as_types:
            self.get_psi(at)
        self.write_data()
   
    @staticmethod
    def get_ioe_event(gtf, exon_len):
        events = ['SE', "SS", "MX", "RI", 'FL']
        ioe_dir = ioe_event(gtf, events, exon_len)
        return ioe_dir
    
    def parse_meta(self, meta):
        self.tpm = {}
        tpm_col = 4
        self.meta_file = meta
        with open(meta, "r") as f:
            try:
                meta_dic = json.load(f)
            except json.decoder.JSONDecodeError:
                print("ERROR: JSON file is excepted!")
                sys.exit()
            self.meta = meta_dic
        read_tpm = partial(self.read_tpm, tpm_col=tpm_col)
        for gn, gdic in meta_dic.items():
            control_tpms = [read_tpm(sp["path"], sp["name"]) for sp in gdic["control"]["samples"]]
            treatment_tpms = [read_tpm(sp["path"], sp["name"]) for sp in gdic["treatment"]["samples"]]
            control_tpm = pd.concat(control_tpms, axis=1)
            treatment_tpm = pd.concat(treatment_tpms, axis=1)
            self.tpm[gn] = [control_tpm, treatment_tpm]

    def get_psi(self, as_type):
        self.psi.setdefault(as_type, {})

        ioe_file = Path(self.ioe_dir) / f"annotation_{as_type}_strict.ioe"
        ioe_df = pd.read_csv(ioe_file, sep="\t")        
        for gn, (tpm1, tpm2) in self.tpm.items():
            _get_psi1 = partial(self._get_psi, tpm=tpm1)
            _get_psi2 = partial(self._get_psi, tpm=tpm2)
            self._get_psi(ioe_df.iloc[1, :], tpm1)

            psi1 = ioe_df.apply(_get_psi1, axis=1)
            psi2 = ioe_df.apply(_get_psi2, axis=1)

            psi1_df = pd.DataFrame(psi1.tolist(), 
                        columns=tpm1.columns, index=ioe_df["event_id"])
            psi2_df = pd.DataFrame(psi2.tolist(), 
                        columns=tpm2.columns, index=ioe_df["event_id"])
            self.psi[as_type][gn] = [psi1_df, psi2_df]

    @staticmethod
    def _get_psi(event_row, tpm):
        alter_tpt = event_row["alternative_transcripts"].split(",")
        total_tpt = event_row["total_transcripts"].split(",")
        al_tpts_val = [i for i in alter_tpt if i in tpm.index]
        all_tpts_val = [i for i in total_tpt if i in tpm.index]

        if len(all_tpts_val) == 0:
            psi_ls = ["nan" for _ in tpm.columns]
        elif len(al_tpts_val) == 0 and len(all_tpts_val) > 0:
            psi_ls = [0 for _ in tpm.columns]
        else:
            get_val = lambda tid, colid: tpm.loc[tid, colid]
            psi_ls = []
            for colid in tpm.columns:
                get_val = lambda tid: tpm.loc[tid, colid]
                al_tpts_abundance = sum(map(get_val, al_tpts_val))
                all_tpts_abundance = sum(map(get_val, all_tpts_val))
                if all_tpts_abundance < 0.001:
                    psi_ls.append("nan")
                    continue
                try:
                    psi = al_tpts_abundance / all_tpts_abundance
                except ZeroDivisionError:
                    psi = "nan"
                psi_ls.append(psi)
        return psi_ls


    @staticmethod
    def read_tpm(file, colname, tpm_col):
        quant_df = pd.read_csv(file, sep = "\t")
        tpm_df = pd.DataFrame({colname: quant_df.iloc[:, tpm_col-1]})
        tpm_df.index = quant_df.iloc[:, 0]
        return tpm_df

    def mk_outdir(self, outdir):
        outdir = Path(outdir)
        outdir.mkdir(exist_ok=True)
        tpm_dir = outdir / "tpm"
        tpm_dir.mkdir(exist_ok=True)
        psi_dir = outdir / "psi"
        psi_dir.mkdir(exist_ok=True)
        dpsi_dir =outdir / "dpsi"
        dpsi_dir.mkdir(exist_ok=True)

        self.outdir = outdir
        self.tpm_dir = tpm_dir
        self.psi_dir = psi_dir
        self.dpsi_dir = dpsi_dir

    def update_meta(self):
        meta = self.meta
        for as_type, group_dic in self.psi_files.items():
           
            for gn, (psi1, psi2) in group_dic.items():
                meta[gn].setdefault("dpsi", {})
                meta[gn]["control"].setdefault("psi", {})
                meta[gn]["control"]["psi"][as_type] = str(psi1)
                meta[gn]["treatment"].setdefault("psi", {})
                meta[gn]["treatment"]["psi"][as_type] = str(psi2)
                meta[gn]["control"]["tpm"] = str(self.tpm_files[gn][0])
                meta[gn]["treatment"]["tpm"] = str(self.tpm_files[gn][1])
                meta[gn]["dpsi"][as_type] = str(self.dpsi_files[as_type][gn])

        with open(self.meta_file, "w") as f:
            json.dump(meta, f, indent=4)

    def write_data(self):
        tpm_dir = self.tpm_dir
        psi_dir = self.psi_dir
        self.tpm_files = {}
        self.psi_files = {}
        for gn, (tpm1, tpm2) in self.tpm.items():
            tpm1_file = tpm_dir / f"{gn}_c1.tpm"
            tpm2_file = tpm_dir / f"{gn}_c2.tpm"
            tpm1.to_csv(tpm1_file, sep="\t")
            tpm2.to_csv(tpm2_file, sep="\t")
            self.tpm_files[gn] = [tpm1_file, tpm2_file]

        for as_type, group_dic in self.psi.items():
            self.psi_files.setdefault(as_type, {})
            for gn, (psi1, psi2) in group_dic.items():
                psi1_file = psi_dir / f"{gn}_{as_type}_c1.psi"
                psi2_file = psi_dir / f"{gn}_{as_type}_c2.psi"
                psi1.to_csv(psi1_file, sep="\t")
                psi2.to_csv(psi2_file, sep="\t")
                self.psi_files[as_type][gn] = [psi1_file, psi2_file]

    #TODO 
    def ds(self, method):
        import shutil
        try:
            shutil.copytree(self.ioe_dir, self.outdir / "ref")
        except FileExistsError as e:
            print(e)
            # print(f"please rm {self.outdir}/ref directory")
        self.method = method
        self.dpsi_files = {}
        from .suppa.lib.diff_tools import multiple_conditions_analysis
        mca = multiple_conditions_analysis

        for as_type, group_dic in self.psi_files.items():
            self.dpsi_files.setdefault(as_type, {})
            for gn, psi_files in group_dic.items():
                expr_files = self.tpm_files[gn]
                dpis_file = self.dpsi_dir / f"{gn}_{as_type}"
                ioe_file = self.ioe_dir / f"annotation_{as_type}_strict.ioe"
                mca(method, psi_files, expr_files, ioe_file, 1000, 0, 
                    False, True, 0.05, True, False, False, 0, 0, str(dpis_file))

                self.dpsi_files[as_type][gn] = dpis_file.with_suffix(".dpsi")
        
        self.update_meta()


#TODO: generate psi files according dpsi(+,-)
class SigFilter:
    def __init__(self, dpsi_file, out, dpsi, pval, 
                abs_dpsi, psi_file, fmt) -> None:
        self.dpsi_file = dpsi_file
        self.dpsi = dpsi
        self.pval = pval
        self.abs_dpsi = abs_dpsi 
        self.psi_file = psi_file if psi_file else []
        self.fmt = fmt
        self.sep = "," if fmt == "csv" else "\t"
        self.sig_psi = []
        self.set_out(out)
 
    def filter_dpsi(self):
        dpsi_df = pd.read_csv(self.dpsi_file, sep="\t", index_col=0)
        old_col = dpsi_df.columns
        dpsi_df.columns = ["dpsi", "pval"]
        filter_df = sig_filter(dpsi_df, dpsi=self.dpsi, abs_dpsi=self.abs_dpsi, pval=self.pval)

        if self.abs_dpsi > 0:
            pos_df = filter_df.loc[filter_df["dpsi"] > 0, ]

            neg_df = filter_df.loc[filter_df["dpsi"] < 0, ]
            self.pos_sig_event = pos_df.index
            self.neg_sig_event = neg_df.index
            
            pos_df.columns = old_col
            neg_df.columns = old_col
            pos_df.to_csv(self.pos_sig_dpsi, index=True, sep=self.sep, index_label=False)
            neg_df.to_csv(self.neg_sig_dpsi, index=True, sep=self.sep, index_label=False)

        filter_df.columns = old_col
        filter_df.to_csv(self.sig_dpsi, index=True, sep=self.sep, index_label=False)

        self.sig_event = filter_df.index
    
    def filter_psi(self):
        for i, pf in enumerate(self.psi_file):
            psi = pd.read_csv(pf, sep="\t")
            psi.index = psi["event_id"]
            if len(set(self.sig_event) & set(psi.index)) > 0:
                sig_psi = psi.loc[self.sig_event, ]
                sig_psi.to_csv(self.sig_psi[i], index=False, sep="\t")
    
    def set_out(self, out):
        sig_psi = []
        if out is not None:
            Path(out).mkdir(exist_ok=True)
            sig_dpsi = Path(out) / Path(self.dpsi_file).name
            for i in self.psi_file:
                sig_psi.append(Path(out) / Path(i).name)
        else:
            self.sig_dpsi = Path(self.dpsi_file)
            for i in self.psi_file:
                sig_psi.append(Path(i))  
        self.sig_dpsi = sig_dpsi.with_suffix(".sig.dpsi")
        if self.abs_dpsi >= 0:
            self.pos_sig_dpsi = sig_dpsi.with_suffix(".sig+.dpsi")
            self.neg_sig_dpsi = sig_dpsi.with_suffix(".sig-.dpsi")
            self.pos_sig_psi = []
            self.neg_sig_psi = []
            for i in sig_psi:
                self.sig_psi.append(i.with_suffix(".sig.psi"))
                self.pos_sig_psi.append(i.with_suffix(".sig+.psi"))
                self.neg_sig_psi.append(i.with_suffix(".sig-.psi"))
        if abs(self.dpsi) >= 0:
            for i in sig_psi:
                self.sig_psi.append(i.with_suffix(".sig.psi"))
    def run(self):
        self.filter_dpsi()
        self.filter_psi()



OrgDb_dic = {
     "hs": "org.Hs.eg.db", "mm": "org.Mm.eg.db", "rn": "org.Rn.eg.db",
     "dm": "org.Dm.eg.db", "at": "org.At.tair.db", "sc": "org.Sc.sgd.db",
     "dr": "org.Dr.eg.db", "ce": "org.Ce.eg.db", "bt": "org.Bt.eg.db",
     "ss": "org.Ss.eg.db", "mmu": "org.Mmu.eg.db", "gg": "org.Gg.eg.db", 
     "cf": "org.Cf.eg.db", "eck12": "org.EcK12.eg.db", "xl": "org.Xl.eg.db",
     "pt": "org.Pt.eg.db", "ag": "org.Ag.eg.db", "pf": "org.Pf.plasmo.db",
     "ecsakai": "org.EcSakai.eg.db", "mxanthus": "org.Mxanthus.db"
     }

def select_OrgDb(org):
    return OrgDb_dic.get(org, None)


def check_kegg_RData(org):
    import subprocess
    rscript = Path(__file__).parent / "R" / "dl_keggdata.R"
    info = subprocess.Popen(["Rscript", str(rscript), org])
    info.wait()
    # import datetime
    # date_str = datetime.datetime.now().strftime("%Y-%m-%d")
    # rpath = ".".join([org, date_str, "kegg.RData"])
    # p = Path(rpath)
    # if p.exists() and p.stat().st_size > 10000:
    #     return True
    # else:
    #     return False


def get_coor(event_id, start, end, strand_sp, anchor, upstream_w, downstream_w):
    eid = fl.EventID(event_id)
    if eid.strand == "-" and strand_sp:
        coordinates = eid.coordinates[::-1]
    else:
        coordinates = eid.coordinates
    if not any([start, end]) and not anchor:
        s, e = eid.alter_element_coor
    elif all([start, end]):
        s, e = sorted([coordinates[start-1], coordinates[end-1]])
    elif anchor:
        anchor_coor = coordinates[anchor-1]
        if eid.strand == "-" and strand_sp:
            e = anchor_coor + upstream_w
            s = anchor_coor - downstream_w
        else:
            s = anchor_coor - upstream_w             
            e = anchor_coor + downstream_w
    else:
        print("param error")
        sys.exit()
    return eid.Chr, s, e, eid.gene_id


def get_coor_bed(dpsi_file, out, start, end, strand_sp, anchor, upstream_w, downstream_w):
    wget_coor_coor = partial(get_coor, start=start, end=end, strand_sp=strand_sp,
                    anchor=anchor, upstream_w=upstream_w, downstream_w=downstream_w)
    dpsi_df = pd.read_csv(dpsi_file, sep="\t", index_col=0)
    dpsi_df["event_id"] = dpsi_df.index
    coors = dpsi_df["event_id"].apply(wget_coor_coor)
    coor_df = pd.DataFrame(coors.tolist())
    coor_df.drop_duplicates(inplace=True)
    coor_df.to_csv(out, index=False, header=False, sep="\t")


def get_anchor_coor(event_id, index, sideindex, offset5, offset3, strand_sp):
    eid = fl.EventID(event_id)
    if eid.strand == "-" and strand_sp:
        coordinates = eid.coordinates[::-1]
    else:
        coordinates = eid.coordinates

    if index is None and sideindex is None:
        s, e = eid.alter_element_coor
        anchor =  s + round((e - s) / 2)
    elif index:
        anchor = coordinates[index-1]
    elif sideindex:
        s, e = sorted([coordinates[i-1] for i in sideindex])
        anchor =  s + round((e - s) / 2)

    if eid.strand == "-" and strand_sp:
        new_anchor = anchor + offset5 - offset3
    else:
        new_anchor = anchor - offset5 + offset3

    return eid.Chr, new_anchor, eid.strand


def gen_anchor_bed(dpsi_file, out, index, sideindex, offset5, offset3, strand_sp):
    import pandas as pd
    from functools import partial

    wget_anchor_coor = partial(get_anchor_coor, index=index, 
            sideindex=sideindex, offset5=offset5, offset3=offset3, 
            strand_sp=strand_sp)
    dpsi_df = pd.read_csv(dpsi_file, sep="\t", index_col=0)
    dpsi_df["event_id"] = dpsi_df.index
    coors = dpsi_df["event_id"].apply(wget_anchor_coor)
    coor_df = pd.DataFrame(coors.tolist())
    coor_df.drop_duplicates(inplace=True)
    coor_df.to_csv(out, index=False, header=False, sep="\t")




# -*- coding: utf-8 -*-
"""
Created on Thu May 22 11:24:33 2014

@author: Gael P Alamancos
@email: gael.perez[at]upf.edu
"""

import sys
from argparse import ArgumentParser, RawTextHelpFormatter, SUPPRESS

description = \
    "Description:\n\n" + \
    "This little script takes 1 or more fields from multiple files with\n" + \
    "a common format and at least a common field which can be used as\n" + \
    "a unique identifier.\n" + \
    "If a identifier do not appear in a file, the corresponding field will\n" + \
    "be set up to 0."
parser = ArgumentParser(description = description, formatter_class=RawTextHelpFormatter)
parser.add_argument("-i", "--input-files", nargs="+", required = True,
                    help = "spaced separated list of files to join.")
parser.add_argument("-k", "--key-field", required = True,
                    help = "common field among the input files.")
parser.add_argument("-f", "--fields", nargs="+", required = True,
                    help = "spaced separated list of fields to select.(starting in 1)")
parser.add_argument("-s", "--separator", default = "\t",
                    help = "field separator")
parser.add_argument("-o", "--output-file", required=True,
                    help = "name of the output file.")
parser.add_argument("--no-header", action="store_true",
                    help = "use it if the file has no header.")
args = parser.parse_args()
try:
    #Recovering the parsed options
    filesList = args.input_files
    keyField = int(args.key_field)
    selectedFields = []
    for i in args.fields:
        selectedFields.append(int(i))
    separator = args.separator
    outputFile = args.output_file
    if args.no_header:
        header = False
    else:
        header = True
    #To remove the key-field from the selectedFields if it is there
    for i in range(len(selectedFields)):
        if selectedFields[i] == keyField:
            del selectedFields[i]
     
    dictionary = {} #key = keyField; value=list of fields to join from different files.         
    if header:
        dictionary["header"] = {}
    #BUFFERING INPUT
    #Loop through all the files    
    for inputFile in filesList:
        seen_header = not header
        print("INFO: Reading file: %s" % inputFile)
        lineNumber = 1
        with open(inputFile, 'r') as f:
            for l in f:
                line = l.rstrip('\n').split(separator)
                if header and not seen_header and (l.startswith('# Transcript') or not(l.startswith('#'))):
                    seen_header = True
                    #Storing the header
                    #print "ASS %s" % line
                    #print selectedFields
                    dictionary["header"][inputFile] =  []
                    for field in selectedFields:
                        dictionary["header"][inputFile].append(line[(field-1)])
                    lineNumber += 1
                    continue
                if l.startswith('#'):
                    continue
                #Initialize to an empty list if the key does not exist
                dictionary.setdefault(line[(keyField-1)], {})
                #Load all the fields from the file
                dictionary[line[(keyField-1)]][inputFile] = []
                for field in selectedFields:
                    dictionary[line[(keyField-1)]][inputFile].append(
                        str(line[(field-1)]))
                lineNumber += 1
        print("INFO: File %s closed." % inputFile)
    #Generating new header
    outputHeader = []
    for inputFile in filesList:
        #aux2 = os.path.basename(inputFile)
        if header:
            aux1 = inputFile.split("/")
            aux1.pop()
            aux2 = aux1.pop()
            #Add file+fields to the header
            for x in dictionary["header"][inputFile]:
                outputHeader.append(aux2)
                #outputHeader.append(aux2 + "_" + str(x))
        else:
            #Add file + num to the header
            aux1 = inputFile.split("/")
            aux1.pop()
            aux2 = aux1.pop()
            for x in range(1, len(selectedFields) + 1):
                outputHeader.append(aux2)
                #outputHeader.append(aux2 + "_" + str(x))

    #WRITING OUTPUT
    print("INFO: Writing output to %s" % outputFile)
    f = open(outputFile, 'w')
    #Writing the header in the output file
    f.write("\t".join(outputHeader) + "\n")
    #Looping through all the unique IDs avoiding the "header" .
    for key, value in [(x,y) for x, y in dictionary.items() if x != "header"]:

        line = []
        line.append(key)    #Adding the common id
        for inputFile in filesList:
            #If the unique id exists in inputFile
            if value.get(inputFile): 
                line += value.get(inputFile)
            #If the unique id does not exist in inputFile
            else:
                #Setting up to 0 all the selectedFields of inputfile
                for x in range(len(selectedFields)):
                    line.append(0)
        #Writing the line in the output
        f.write("\t".join(line) + "\n")
    f.close()
    print("INFO: %s closed." % outputFile)
    print("INFO: Process ended succesfully.")
except:
    print("ERROR: %s" % err)
    sys.exit(1)



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 10 14:57:05 2014

@authors: Gael P Alamancos and Miha Skalic
@email: gael.perez[at]upf.edu, miha.skalic[at]gmail.com
"""

import sys
import logging
from argparse import ArgumentParser, RawTextHelpFormatter
from lib.tools import *
from lib.gtf_store import *
from lib.event import *

# Setting argument parser
# parser = argparse.ArgumentParser()
description = \
    "Description:\n\n" + \
    "This tool reads an annotation file and generates different alternative\n" + \
    "splicing(AS) events depending on the user's choice.\n" + \
    "It outputs a \"gtf\" file to load on genome browser to visualize the \n" + \
    "events as well as an \"ioe\" file to use for the further calculation of\n" + \
    "the PSI values of each possible event conformation."
parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter,
                        add_help=False)
parser.add_argument("-i", "--input-file", help="specify input file",
                    required=True)
parser.add_argument("-o", "--output-file", help="specify output path and" +
                    " name without extension", required=True)

parser.add_argument("-e", "--event-type", nargs='+', required=False, choices=["SE", "SS", "MX", "RI", "FL"],
                    help="list of events to analyze. "
                    "(space separated)\n\n"
                    "Options:\n"
                    "\tSE -- Skipping Exon\n"
                    "\tSS -- Alternative Splice Site (5'/3')\n"
                    "\tMX -- Mutually Exclusive Exon\n"
                    "\tRI -- Retained Intron\n"
                    "\tFL -- Alternative First/Last Exon\n")
parser.add_argument("-b", "--boundary", choices=["S", "V"], default="S", help="Boundary type."
                    "Options:\n"
                    "\tS -- Strict (Default)\n"
                    "\tV -- Variable\n")
parser.add_argument("-t", "--threshold", default=10, type=int,
                    help="Variability treshold. In case of strict boundaries this argument is ignored" +
                    "(Default: 10nt).")
parser.add_argument("-p", "--pool-genes", action="store_true",
                    help="pool together overlapping genes.")
parser.add_argument("-l", "--exon-length", default=100, type=int,
                    help="length of the exons for its visualization. " +
                    "(Default: 100nt)")
parser.add_argument("-m", "--mode", default="INFO",
                    help="to choose from DEBUG, INFO, WARNING, ERROR and CRITICAL")
parser.add_argument('-f', '--format',
                    dest="format",
                    action="store",
                    required=True,
                    choices=['ioe', 'ioi'],
                    help="Format of the annotation file. Required.")


def main():

    args = parser.parse_args()

    # Parsing arguments
    mode = "logging." + args.mode

    # Setting logging preferences
    logger = logging.getLogger(__name__)
    logger.setLevel(eval(mode))

    # Setting the level of the loggers in lib
    setToolsLoggerLevel(mode)

    exon_length = int(args.exon_length)

    my_genome = Genome()

    # Check if event list exist only if format is IOE
    if args.format == "ioe" and args.event_type == None:
        logger.info("No event type found. It is necessary to determine at least a type of event to create an IOE file.")
        exit(1)

    logger.info("Reading input data.")
    fetched_exons = gtf_reader(args.input_file, logger)

    # Check for empy sequence
    if len(fetched_exons) == 0:
        logger.info("No exons found. Check format and content of your GTF file.")
        exit(1)

    for exon_meta in fetched_exons:
        my_genome.add_to_genes(exon_meta)

    my_genome.sort_transcripts()

    if args.pool_genes:
        my_genome.split_genes()
        logger.info("Pooling genes")
        my_genome.poll_genes()

    if args.format == "ioi":
        ioi_writer(my_genome, args.output_file)
    else:
        make_events(args.event_type, my_genome, args.input_file, args.output_file, exon_length, logger,
                b_type=args.boundary, th=args.threshold)
    logger.info("Done")
    
if __name__ == '__main__':
    main()



from setuptools import setup
from setuptools import find_packages

setup(
    name="astk",
    version="0.1.1",
    url='https://github.com/huang-sh/astk/',
    author='huangsh',
    author_email='hsh-me@outlook.com',
    packages =  find_packages(),
    package_dir = {"astk": "astk"},
    include_package_data=True,
    install_requires=[
        "click>=8.0.0",
        "pandas>=1.3.0",
        "matplotlib>=3.4.0",
        "scikit-learn>=0.24.0",
        "statsmodels>= 0.12.0"
    ],
    entry_points="""
        [console_scripts]
        astk=astk.astk:cli
    """,
)


# -*- coding: utf-8 -*-
"""
Created on Wed May 25 04:20:00 CEST 2016

@authors: Juan C Entizne, Juan L. Trincado
@email: juancarlos.entizne01[at]estudiant.upf.edu,
        juanluis.trincado[at]upf.edu
"""

import os
import logging
from lib.cluster_tools import cluster_analysis
from argparse import ArgumentParser, RawTextHelpFormatter


description = \
    "Description:\n\n" + \
    "This tool cluster events that change significantly in at least one pair of conditions, across multiple conditions.\n" + \
    "This tool takes as input the .dpsi and .psivec files generate by SUPPA differentialAnalysis method\n" + \
    "and generates a .clustvec on which the events has been tagged according to their cluster membership\n"


parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter,
                        add_help=False)

parser.add_argument('-d', '--dpsi',
                    dest="dpsi",
                    nargs=1,
                    action="store",
                    help="Input file of delta-PSI values (.dpsi format)")

parser.add_argument('-p', '--psivec',
                    dest="psivec",
                    nargs=1,
                    action="store",
                    help="Input file with PSI values (.psivec format)")

parser.add_argument('-st', '--sig-threshold',
                    dest="sig_threshold",
                    action="store",
                    type=float,
                    default=0.05,
                    help="P-value cut-off for significant events. (Default: 0.05).")

parser.add_argument('-dt', '--dpsi-threshold',
                    dest="dpsi_threshold",
                    action="store",
                    type=float,
                    default=0.05,
                    help="Lower-bound for the absolute delta PSI value to cluster. (Default: 0.05).")

parser.add_argument('-e', '--eps',
                    dest="eps",
                    action="store",
                    type=float,
                    default=0.05,
                    help="Maximum (Euclidean) distance (between 0 and 1) to consider two events as members of "
                         "the same cluster. (Default: 0.05).")

parser.add_argument('-s', '--separation',
                    action="store",
                    type=float,
                    default=0,
                    help="Minimum separation for considering two points in different clusters. (Default: 0).")

parser.add_argument('-n', '--min-pts',
                    dest="minpts",
                    action="store",
                    type=int,
                    default=20,
                    help="Minimum number of events required per cluster. (Default: 20).")

parser.add_argument("-m", "--metric", dest="metric", choices=["euclidean", "manhattan", "cosine"],
                    default="euclidean", help="Distance function to be used."
                                                                              "Options:\n"
                                                                              "\teuclidean (Default),\n"
                                                                              "\tmanhattan,\n"
                                                                              "\tcosine.\n")
parser.add_argument("-c", "--clustering", choices=["OPTICS", "DBSCAN"],
                    default="DBSCAN", help="Clustering method to use."
                                                                              "Options:\n"
                                                                              "\tOPTICS ,\n"
                                                                              "\tDBSCAN (Default).\n")
parser.add_argument('-g', '--groups',
                    dest="indexes",
                    action="store",
                    required=True,
                    type=str,
                    nargs="*",
                    help="Ranges of column numbers specifying the replicates per condition. "
                         "Column numbers have to be continuous, with no overlapping or missing columns between them. "
                         "Ex: 1-3,4-6")

parser.add_argument('-o', '--output',
                    dest="output",
                    action="store",
                    help="Name of the output file.")


def create_path(lst):

    temp_lst = []
    for fl in lst:
        if not os.path.isabs(fl):
            fl_path = os.getcwd()+"/"+fl
            temp_lst.append(fl_path)
        else:
            temp_lst.append(fl)

    return temp_lst


def main():

    args = parser.parse_args()

    # Check if path is absolute, if not the program use the current working path
    dpsi_file = create_path(args.dpsi)
    psivec_file = create_path(args.psivec)

    cluster_analysis(dpsi_file[0], psivec_file[0], args.sig_threshold, args.dpsi_threshold, args.eps, args.minpts,
                     args.metric,args.indexes[0], args.clustering, args.separation, args.output)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Fri May 23 10:17:33 2014

@author: Miha Skalic
@email: miha.skalic[at]gmail.com
"""


import sys
import logging
from argparse import ArgumentParser, RawTextHelpFormatter
from lib.tools import *
from lib.gtf_store import *


description = \
    "Description:\n\n" + \
    "This tool calculates the PSI (Percentatge Splice In) for the different\n" + \
    "transcripts of a gene.\n" + \
    "It reads a gtf to get transcript-gene relationship and an expression file\n" + \
    "of the different transcripts\n" 
    
parser = ArgumentParser(description=description, formatter_class=RawTextHelpFormatter,
                        add_help=False)
parser.add_argument("-g", "--gtf-file", help="Input gtf file",
                    required=True)
parser.add_argument("-e", "--expression-file", required=True,
                    help="Input expression file")
parser.add_argument("-o", "--output-file", required=True,
                    help="Path and name of the ouput file")
parser.add_argument("-m", "--mode", default="INFO",
                    help="to choose from DEBUG, INFO, WARNING, ERROR and CRITICAL")


def expression_reader(exp_file):
    """
    Reads in expression file and returns dict
    of transcript expressions and first line.
    """
    if not os.path.isfile(exp_file):
        sys.stderr.write("Expression file does not exist. Quiting\n")
        exit(1)

    expressions = {}
    with open(exp_file, 'r') as handle:
        first_line = nextel(handle).strip()
        for line in handle:
            line = line.strip().split('\t')
            expressions[line[0]] = [float(xp) for xp in line[1:]]
    return expressions, first_line


def expression_writer(genomeinfo, expressions, firstline, output_file):
    """
    Function to write perIsoform inclusion
    """
    output_file += '_isoform.psi'
    entriesnumber = len(expressions[nextel(expressions.__iter__())])
    with open(output_file, 'w') as handle:
        handle.write(firstline + '\n')
        for gene, _, _ in genomeinfo:
            expr_sum = [0 for _ in range(entriesnumber)]

            # collect expression
            for transcript in gene.sortedTranscripts:
                if transcript not in expressions:
                    logger.info(('Expression for transcript "{}" not found. '
                                 'Ignoring it in calculation.').format(transcript))
                else:
                    expr_sum = list(map(lambda exp_pair: exp_pair[0] + exp_pair[1], zip(expr_sum, expressions[transcript])))

            # calculate expression
            if 0 in expr_sum:
                logger.debug('Gene "{}" has at least one replicate with 0 expression.'.format(gene.name))
                expr_sum = [y if y else float('NaN') for y in expr_sum]

            for transcript in gene.sortedTranscripts:
                if transcript not in expressions:
                    continue
                t_exp = map(lambda exp_pair: exp_pair[1] / exp_pair[0], zip(expr_sum, expressions[transcript]))
                handle.write('{};{}\t{}\n'.format(gene.name, transcript,
                                                  '\t'.join([str(exp_val) for exp_val in t_exp])))


def main(): 
    args = parser.parse_args()

    #Parsing arguments
    mode = "logging." + args.mode

    #Setting logging preferences
    logger = logging.getLogger(__name__)
    logger.setLevel(eval(mode))

    #Setting the level of the loggers in lib
    setToolsLoggerLevel(mode)

    #PREPAIRING GTF
    my_genome = Genome()
    logger.info("Reading GTF data.")
    fetched_exons = gtf_reader(args.gtf_file, logger)

    # Check for empy sequences
    if len(fetched_exons) == 0:
        logger.info("No exons found. Check format and content of your GTF file.")
        exit(1)

    for exon_meta in fetched_exons:
        my_genome.add_to_genes(exon_meta)

    # split non overlapping genes
    my_genome.sort_transcripts()
    my_genome.split_genes()

    logger.info("Reading Expression data.")
    trans_expres, sample_names = expression_reader(args.expression_file)
    if not trans_expres:
        logger.info("No expressions found. Check format and content of your expression file.")
        exit(1)

    # Calculate and write output
    logger.info("Calculating inclusion and generating output.")
    expression_writer(my_genome, trans_expres, sample_names, args.output_file)


if __name__ == '__main__':
    main()
